# This file is part of allegedb, an object relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""allegedb's special implementations of the NetworkX graph objects"""
from abc import ABC

import networkx
from networkx.exception import NetworkXError
from collections import defaultdict
from collections.abc import MutableMapping
from .wrap import MutableMappingUnwrapper


class EntityCollisionError(ValueError):
	"""For when there's a discrepancy between the kind of entity you're creating and the one by the same name"""


def getatt(attribute_name):
	"""An easy way to make an alias"""
	from operator import attrgetter
	ret = property(attrgetter(attribute_name))
	ret.__doc__ = "Alias to `{}`".format(attribute_name)
	return ret


def convert_to_networkx_graph(data, create_using=None, multigraph_input=False):
	"""Convert an AllegedGraph to the corresponding NetworkX graph type."""
	if isinstance(data, DiGraph):
		result = networkx.convert.from_dict_of_dicts(
			data.adj,
			create_using=create_using,
			multigraph_input=data.is_multigraph())
		result.graph = dict(data.graph)
		result.node = {k: dict(v) for k, v in data.node.items()}
		return result
	return networkx.convert.to_networkx_graph(data, create_using,
												multigraph_input)


_alleged_receivers = defaultdict(list)


class AllegedMapping(MutableMappingUnwrapper, ABC):
	"""Common amenities for mappings"""
	__slots__ = ()

	def connect(self, func):
		"""Arrange to call this function whenever something changes here.

        The arguments will be this object, the key changed, and the value set.

        """
		l = _alleged_receivers[id(self)]
		if func not in l:
			l.append(func)

	def disconnect(self, func):
		"""No longer call the function when something changes here."""
		if id(self) not in _alleged_receivers:
			return
		l = _alleged_receivers[id(self)]
		try:
			l.remove(func)
		except ValueError:
			return
		if not l:
			del _alleged_receivers[id(self)]

	def send(self, sender, **kwargs):
		"""Internal. Call connected functions."""
		if id(self) not in _alleged_receivers:
			return
		for func in _alleged_receivers[id(self)]:
			func(sender, **kwargs)

	def clear(self):
		"""Delete everything"""
		for k in list(self.keys()):
			del self[k]


class AbstractEntityMapping(AllegedMapping, ABC):
	__slots__ = ()
	db: 'ORM'

	def _get_cache(self, key, branch, turn, tick):
		raise NotImplementedError

	def _get_cache_now(self, key):
		return self._get_cache(key, *self.db._btt())

	def _cache_contains(self, key, branch, turn, tick):
		raise NotImplementedError

	def _set_db(self, key, branch, turn, tick, value):
		"""Set a value for a key in the database (not the cache)."""
		raise NotImplementedError

	def _set_cache(self, key, branch, turn, tick, value):
		raise NotImplementedError

	def _set_cache_now(self, key, value):
		branch, turn, tick = self.db._nbtt()
		self._set_cache(key, branch, turn, tick, value)

	def _del_db(self, key, branch, turn, tick):
		"""Delete a key from the database (not the cache)."""
		self._set_db(key, branch, turn, tick, None)

	def _del_cache(self, key, branch, turn, tick):
		self._set_cache(key, branch, turn, tick, None)

	def __getitem__(self, key):
		"""If key is 'graph', return myself as a dict, else get the present
        value of the key and return that

        """

		def wrapval(v):
			from functools import partial
			from .wrap import DictWrapper, ListWrapper, SetWrapper
			if isinstance(v, list):
				return ListWrapper(partial(self._get_cache_now, key),
									partial(self._set_cache_now, key), self,
									key)
			elif isinstance(v, dict):
				return DictWrapper(partial(self._get_cache_now, key),
									partial(self._set_cache_now, key), self,
									key)
			elif isinstance(v, set):
				return SetWrapper(partial(self._get_cache_now, key),
									partial(self._set_cache_now, key), self,
									key)
			else:
				return v

		return wrapval(self._get_cache_now(key))

	def __contains__(self, item):
		return item == 'name' or self._cache_contains(item, *self.db._btt())

	def __setitem__(self, key, value):
		"""Set key=value at the present branch and revision"""
		if value is None:
			raise ValueError(
				"allegedb uses None to indicate that a key's been deleted")
		branch, turn, tick = self.db._nbtt()
		try:
			if self._get_cache(key, branch, turn, tick) != value:
				self._set_cache(key, branch, turn, tick, value)
		except KeyError:
			self._set_cache(key, branch, turn, tick, value)
		self._set_db(key, branch, turn, tick, value)
		self.send(self, key=key, value=value)

	def __delitem__(self, key):
		branch, turn, tick = self.db._nbtt()
		self._del_cache(key, branch, turn, tick)
		self._del_db(key, branch, turn, tick)
		self.send(self, key=key, value=None)


class GraphMapping(AbstractEntityMapping):
	"""Mapping for graph attributes"""
	__slots__ = ('graph', 'db', '_iter_stuff', '_cache_contains_stuff',
					'_len_stuff', '_get_stuff', '_set_db_stuff',
					'_set_cache_stuff', '_del_db_stuff', '_get_cache_stuff')

	def __init__(self, graph):
		super().__init__()
		self.graph = graph
		self.db = db = graph.db
		btt = db._btt
		graph_val_cache = db._graph_val_cache
		graphn = graph.name
		self._iter_stuff = (graph_val_cache.iter_entity_keys, graphn, btt)
		self._cache_contains_stuff = (graph_val_cache.contains_key, graphn)
		self._len_stuff = (graph_val_cache.count_entities, graphn, btt)
		self._get_stuff = (self._get_cache, btt)
		graph_val_set = db.query.graph_val_set
		self._set_db_stuff = (graph_val_set, graphn)
		self._set_cache_stuff = (graph_val_cache.store, graphn)
		self._del_db_stuff = (graph_val_set, graphn)
		self._get_cache_stuff = (graph_val_cache.retrieve, graphn)

	def __iter__(self):
		iter_entity_keys, graphn, btt = self._iter_stuff
		yield 'name'
		yield from iter_entity_keys(graphn, *btt())

	def _cache_contains(self, key, branch, turn, tick):
		contains_key, graphn = self._cache_contains_stuff
		return contains_key(graphn, key, branch, turn, tick)

	def __len__(self):
		count_entities, graphn, btt = self._len_stuff
		return 1 + count_entities(graphn, *btt())

	def __getitem__(self, item):
		if item == 'name':
			return self.graph.name
		return super().__getitem__(item)

	def __setitem__(self, key, value):
		if key == 'name':
			raise KeyError("name cannot be changed after creation")
		super().__setitem__(key, value)

	def _get_cache(self, key, branch, turn, tick):
		retrieve, graphn = self._get_cache_stuff
		return retrieve(graphn, key, branch, turn, tick)

	def _get(self, key):
		get_cache, btt = self._get_stuff
		return get_cache(key, *btt())

	def _set_db(self, key, branch, turn, tick, value):
		graph_val_set, graphn = self._set_db_stuff
		graph_val_set(graphn, key, branch, turn, tick, value)

	def _set_cache(self, key, branch, turn, tick, value):
		store, graphn = self._set_cache_stuff
		store(graphn, key, branch, turn, tick, value)

	def _del_db(self, key, branch, turn, tick):
		graph_val_set, graphn = self._del_db_stuff
		graph_val_set(graphn, key, branch, turn, tick, None)

	def clear(self):
		keys = set(self.keys())
		keys.remove('name')
		for k in keys:
			del self[k]

	def unwrap(self):
		return {
			k: v.unwrap()
			if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap') else v
			for (k, v) in self.items()
		}

	def __eq__(self, other):
		if hasattr(other, 'unwrap'):
			other = other.unwrap()
		other = other.copy()
		me = self.unwrap().copy()
		if 'name' not in other:
			del me['name']
		return me == other


class Node(AbstractEntityMapping):
	"""Mapping for node attributes"""
	__slots__ = ('graph', 'node', 'db', '__weakref__', '_iter_stuff',
					'_cache_contains_stuff', '_len_stuff', '_get_cache_stuff',
					'_set_db_stuff', '_set_cache_stuff')

	def _validate_node_type(self):
		return True

	def __new__(cls, graph, node, clobber=False):
		gnn = (graph.name, node)
		nobjs = graph.db._node_objs
		if clobber and gnn in nobjs:
			del nobjs[gnn]
		elif gnn in nobjs:
			ret = nobjs[gnn]
			if not isinstance(ret, cls):
				raise EntityCollisionError(
					"Already have node {} in graph {}, but it's of class {}".
					format(node, graph.name,
							type(ret).__name__))
			return ret
		return super(Node, cls).__new__(cls)

	def __init__(self, graph, node, clobber=False):
		"""Store name and graph"""
		super().__init__()
		self.graph = graph
		self.node = node
		self.db = db = graph.db
		node_val_cache = db._node_val_cache
		graphn = graph.name
		btt = db._btt
		self._iter_stuff = (node_val_cache.iter_entity_keys, graphn, node, btt)
		self._cache_contains_stuff = (node_val_cache.contains_key, graphn,
										node)
		self._len_stuff = (node_val_cache.count_entity_keys, graphn, node, btt)
		self._get_cache_stuff = (node_val_cache.retrieve, graphn, node)
		self._set_db_stuff = (db.query.node_val_set, graphn, node)
		self._set_cache_stuff = (db._node_val_cache.store, graphn, node)

	def __repr__(self):
		return "{}(graph={}, node={})".format(self.__class__.__name__,
												self.graph, self.node)

	def __str__(self):
		return "{}(graph={}, node={}, data={})".format(self.__class__.__name__,
														self.graph, self.node,
														repr(dict(self)))

	def __iter__(self):
		iter_entity_keys, graphn, node, btt = self._iter_stuff
		return iter_entity_keys(graphn, node, *btt())

	def _cache_contains(self, key, branch, turn, tick):
		contains_key, graphn, node = self._cache_contains_stuff
		return contains_key(graphn, node, key, branch, turn, tick)

	def __len__(self):
		count_entity_keys, graphn, node, btt = self._len_stuff
		return count_entity_keys(graphn, node, *btt())

	def _get_cache(self, key, branch, turn, tick):
		retrieve, graphn, node = self._get_cache_stuff
		return retrieve(graphn, node, key, branch, turn, tick)

	def _set_db(self, key, branch, turn, tick, value):
		node_val_set, graphn, node = self._set_db_stuff
		node_val_set(graphn, node, key, branch, turn, tick, value)

	def _set_cache(self, key, branch, turn, tick, value):
		store, graphn, node = self._set_cache_stuff
		store(graphn, node, key, branch, turn, tick, value)

	def __eq__(self, other):
		if not isinstance(other, type(self)):
			return False
		return dict(self) == dict(other)


class Edge(AbstractEntityMapping):
	"""Mapping for edge attributes"""
	__slots__ = ('graph', 'orig', 'dest', 'idx', 'db', '__weakref__',
					'_iter_stuff', '_cache_contains_stuff', '_len_stuff',
					'_get_cache_stuff', '_set_db_stuff', '_set_cache_stuff')

	set_db_time = set_cache_time = 0

	def __new__(cls, graph, orig, dest, idx=0):
		gnodi = (graph.name, orig, dest, idx)
		edgeobjs = graph.db._edge_objs
		if gnodi in edgeobjs:
			ret = edgeobjs[gnodi]
			if not isinstance(ret, cls):
				raise EntityCollisionError(
					"Already have an edge {}->{}[{}] in graph {}, but of class {}"
					.format(orig, dest, idx, graph.name, type(ret)))
			return ret
		return super(Edge, cls).__new__(cls)

	def __init__(self, graph, orig, dest, idx=0):
		"""Store the graph, the names of the nodes, and the index.

        For non-multigraphs the index is always 0.

        """
		super().__init__()
		self.graph = graph
		self.db = db = graph.db
		self.orig = orig
		self.dest = dest
		self.idx = idx
		edge_val_cache = db._edge_val_cache
		graphn = graph.name
		btt = db._btt
		self._iter_stuff = (edge_val_cache.iter_entity_keys, graphn, orig,
							dest, idx, btt)
		self._cache_contains_stuff = (edge_val_cache.contains_key, graphn,
										orig, dest, idx)
		self._len_stuff = (edge_val_cache.count_entity_keys, graphn, orig,
							dest, idx, btt)
		self._get_cache_stuff = (edge_val_cache.retrieve, graphn, orig, dest,
									idx)
		self._set_db_stuff = (db.query.edge_val_set, graphn, orig, dest, idx)
		self._set_cache_stuff = (edge_val_cache.store, graphn, orig, dest, idx)

	def __repr__(self):
		return "<{} in graph {} from {} to {} containing {}>".format(
			self.__class__.__name__, self.graph, self.orig, self.dest,
			dict(self))

	def __str__(self):
		return str(dict(self))

	def __iter__(self):
		iter_entity_keys, graphn, orig, dest, idx, btt = self._iter_stuff
		return iter_entity_keys(graphn, orig, dest, idx, *btt())

	def _cache_contains(self, key, branch, turn, tick):
		contains_key, graphn, orig, dest, idx = self._cache_contains_stuff
		return contains_key(graphn, orig, dest, idx, key, branch, turn, tick)

	def __len__(self):
		count_entity_keys, graphn, orig, dest, idx, btt = self._len_stuff
		return count_entity_keys(graphn, orig, dest, idx, *btt())

	def _get_cache(self, key, branch, turn, tick):
		retrieve, graphn, orig, dest, idx = self._get_cache_stuff
		return retrieve(graphn, orig, dest, idx, key, branch, turn, tick)

	def _set_db(self, key, branch, turn, tick, value):
		edge_val_set, graphn, orig, dest, idx = self._set_db_stuff
		edge_val_set(graphn, orig, dest, idx, key, branch, turn, tick, value)

	def _set_cache(self, key, branch, turn, tick, value):
		store, graphn, orig, dest, idx = self._set_cache_stuff
		store(graphn, orig, dest, idx, key, branch, turn, tick, value)


class GraphNodeMapping(AllegedMapping):
	"""Mapping for nodes in a graph"""
	__slots__ = ('graph', )

	db = getatt('graph.db')
	"""Alias to ``self.graph.db``"""

	def __init__(self, graph):
		super().__init__()
		self.graph = graph

	def __iter__(self):
		"""Iterate over the names of the nodes"""
		return self.db._nodes_cache.iter_entities(self.graph.name,
													*self.db._btt())

	def __eq__(self, other):
		from collections.abc import Mapping
		if not isinstance(other, Mapping):
			return NotImplemented
		if self.keys() != other.keys():
			return False
		for k in self.keys():
			me = self[k]
			you = other[k]
			if hasattr(me, 'unwrap') and not hasattr(me, 'no_unwrap'):
				me = me.unwrap()
			if hasattr(you, 'unwrap') and not hasattr(you, 'no_unwrap'):
				you = you.unwrap()
			if me != you:
				return False
		else:
			return True

	def __contains__(self, node):
		"""Return whether the node exists presently"""
		return self.db._nodes_cache.contains_entity(self.graph.name, node,
													*self.db._btt())

	def __len__(self):
		"""How many nodes exist right now?"""
		return self.db._nodes_cache.count_entities(self.graph.name,
													*self.db._btt())

	def __getitem__(self, node):
		"""If the node exists at present, return it, else throw KeyError"""
		if node not in self:
			raise KeyError
		return self.db._get_node(self.graph, node)

	def __setitem__(self, node, dikt):
		"""Only accept dict-like values for assignment. These are taken to be
        dicts of node attributes, and so, a new GraphNodeMapping.Node
        is made with them, perhaps clearing out the one already there.

        """
		created = False
		db = self.db
		graph = self.graph
		gname = graph.name
		if not db._node_exists(gname, node):
			created = True
			db._exist_node(gname, node, True)
		n = db._get_node(graph, node)
		n.clear()
		n.update(dikt)
		if created:
			self.send(self, node_name=node, exists=True)

	def __delitem__(self, node):
		"""Indicate that the given node no longer exists"""
		if node not in self:
			raise KeyError("No such node")
		for succ in self.graph.adj[node]:
			del self.graph.adj[node][succ]
		for pred in self.graph.pred[node]:
			del self.graph.pred[node][pred]
		branch, turn, tick = self.db._nbtt()
		self.db.query.exist_node(self.graph.name, node, branch, turn, tick,
									False)
		self.db._nodes_cache.store(self.graph.name, node, branch, turn, tick,
									False)
		key = (self.graph.name, node)
		if node in self.db._node_objs:
			del self.db._node_objs[key]
		self.send(self, node_name=node, exists=False)


class GraphEdgeMapping(AllegedMapping):
	"""Provides an adjacency mapping and possibly a predecessor mapping
    for a graph.

    """
	__slots__ = ('graph', '_cache')

	db = getatt('graph.db')
	"""Alias to ``self.graph.db``"""

	def __init__(self, graph):
		super().__init__()
		self.graph = graph
		self._cache = {}

	def __eq__(self, other):
		"""Compare dictified versions of the edge mappings within me.

        As I serve custom Predecessor or Successor classes, which
        themselves serve the custom Edge class, I wouldn't normally be
        comparable to a networkx adjacency dictionary. Converting
        myself and the other argument to dicts allows the comparison
        to work anyway.

        """
		if not hasattr(other, 'keys'):
			return False
		if self.keys() != other.keys():
			return False
		for k in self.keys():
			if dict(self[k]) != dict(other[k]):
				return False
		return True

	def __iter__(self):
		return iter(self.graph.node)


class AbstractSuccessors(GraphEdgeMapping):
	__slots__ = ('graph', 'container', 'orig', '_cache')

	db = getatt('graph.db')
	"""Alias to ``self.graph.db``"""

	def _order_nodes(self, node):
		raise NotImplementedError

	def __init__(self, container, orig):
		"""Store container and node"""
		super().__init__(container.graph)
		self.container = container
		self.orig = orig

	def __iter__(self):
		"""Iterate over node IDs that have an edge with my orig"""
		return self.db._edges_cache.iter_successors(self.graph.name, self.orig,
													*self.db._btt())

	def __contains__(self, dest):
		"""Is there an edge leading to ``dest`` at the moment?"""
		orig, dest = self._order_nodes(dest)
		return self.db._edges_cache.has_successor(self.graph.name, orig, dest,
													*self.db._btt())

	def __len__(self):
		"""How many nodes touch an edge shared with my orig?"""
		return self.db._edges_cache.count_successors(self.graph.name,
														self.orig,
														*self.db._btt())

	def _make_edge(self, dest):
		return Edge(self.graph, *self._order_nodes(dest))

	def __getitem__(self, dest):
		"""Get the edge between my orig and the given node"""
		if dest not in self:
			raise KeyError("No edge {}->{}".format(self.orig, dest))
		orig, dest = self._order_nodes(dest)
		return self.db._get_edge(self.graph, orig, dest, 0)

	def __setitem__(self, dest, value):
		"""Set the edge between my orig and the given dest to the given
        value, a mapping.

        """
		real_dest = dest
		orig, dest = self._order_nodes(dest)
		created = dest not in self
		if orig not in self.graph.node:
			self.graph.add_node(orig)
		if dest not in self.graph.node:
			self.graph.add_node(dest)
		branch, turn, tick = self.db._nbtt()
		self.db.query.exist_edge(self.graph.name, orig, dest, 0, branch, turn,
									tick, True)
		self.db._edges_cache.store(self.graph.name, orig, dest, 0, branch,
									turn, tick, True)
		e = self[real_dest]
		e.clear()
		e.update(value)
		if created:
			self.send(self, orig=orig, dest=dest, idx=0, exists=True)

	def __delitem__(self, dest):
		"""Remove the edge between my orig and the given dest"""
		branch, turn, tick = self.db._nbtt()
		orig, dest = self._order_nodes(dest)
		self.db.query.exist_edge(self.graph.name, orig, dest, 0, branch, turn,
									tick, False)
		self.db._edges_cache.store(self.graph.name, orig, dest, 0, branch,
									turn, tick, None)
		self.send(self, orig=orig, dest=dest, idx=0, exists=False)

	def __repr__(self):
		cls = self.__class__
		return "<{}.{} object containing {}>".format(cls.__module__,
														cls.__name__,
														dict(self))

	def clear(self):
		"""Delete every edge with origin at my orig"""
		for dest in list(self):
			del self[dest]


class GraphSuccessorsMapping(GraphEdgeMapping):
	"""Mapping for Successors (itself a MutableMapping)"""
	__slots__ = ('graph', )

	class Successors(AbstractSuccessors):
		__slots__ = ('graph', 'container', 'orig', '_cache')

		def _order_nodes(self, dest):
			if dest < self.orig:
				return (dest, self.orig)
			else:
				return (self.orig, dest)

	def __getitem__(self, orig):
		if orig not in self._cache:
			self._cache[orig] = self.Successors(self, orig)
		return self._cache[orig]

	def __setitem__(self, key, val):
		"""Wipe out any edges presently emanating from orig and replace them
        with those described by val

        """
		if key in self:
			sucs = self[key]
			created = False
		else:
			sucs = self._cache[key] = self.Successors(self, key)
			created = True
			sucs.clear()
		if val:
			sucs.update(val)
		if created:
			self.send(self, key=key, val=val)

	def __delitem__(self, key):
		"""Wipe out edges emanating from orig"""
		self[key].clear()
		del self._cache[key]
		self.send(self, key=key, val=None)

	def __iter__(self):
		return iter(self.graph.node)

	def __len__(self):
		return len(self.graph.node)

	def __contains__(self, key):
		return key in self.graph.node

	def __repr__(self):
		cls = self.__class__
		return "<{}.{} object containing {}>".format(
			cls.__module__, cls.__name__, {
				k: {k2: dict(v2)
					for (k2, v2) in v.items()}
				for (k, v) in self.items()
			})


class DiGraphSuccessorsMapping(GraphSuccessorsMapping):
	__slots__ = ('graph', )

	class Successors(AbstractSuccessors):
		__slots__ = ('graph', 'container', 'orig', '_cache')

		def _order_nodes(self, dest):
			return (self.orig, dest)


class DiGraphPredecessorsMapping(GraphEdgeMapping):
	"""Mapping for Predecessors instances, which map to Edges that end at
    the dest provided to this

    """
	__slots__ = ('graph', )

	def __contains__(self, dest):
		return dest in self.graph.node

	def __getitem__(self, dest):
		"""Return a Predecessors instance for edges ending at the given
        node

        """
		if dest not in self:
			raise KeyError("No edges available")
		if dest not in self._cache:
			self._cache[dest] = self.Predecessors(self, dest)
		return self._cache[dest]

	def __setitem__(self, key, val):
		"""Interpret ``val`` as a mapping of edges that end at ``dest``"""
		created = key not in self
		if key not in self._cache:
			self._cache[key] = self.Predecessors(self, key)
		preds = self._cache[key]
		preds.clear()
		preds.update(val)
		if created:
			self.send(self, key=key, val=val)

	def __delitem__(self, key):
		"""Delete all edges ending at ``dest``"""
		it = self[key]
		it.clear()
		del self._cache[key]
		self.send(self, key=key, val=None)

	def __iter__(self):
		return iter(self.graph.node)

	def __len__(self):
		return len(self.graph.node)

	class Predecessors(GraphEdgeMapping):
		"""Mapping of Edges that end at a particular node"""
		__slots__ = ('graph', 'container', 'dest')

		def __init__(self, container, dest):
			"""Store container and node ID"""
			super().__init__(container.graph)
			self.container = container
			self.dest = dest

		def __iter__(self):
			"""Iterate over the edges that exist at the present (branch, rev)

            """
			return self.db._edges_cache.iter_predecessors(
				self.graph.name, self.dest, *self.db._btt())

		def __contains__(self, orig):
			"""Is there an edge from ``orig`` at the moment?"""
			return self.db._edges_cache.has_predecessor(
				self.graph.name, self.dest, orig, *self.db._btt())

		def __len__(self):
			"""How many edges exist at this rev of this branch?"""
			return self.db._edges_cache.count_predecessors(
				self.graph.name, self.dest, *self.db._btt())

		def _make_edge(self, orig):
			return Edge(self.graph, orig, self.dest)

		def __getitem__(self, orig):
			"""Get the edge from the given node to mine"""
			return self.graph.adj[orig][self.dest]

		def __setitem__(self, orig, value):
			"""Use ``value`` as a mapping of edge attributes, set an edge from the
            given node to mine.

            """
			branch, turn, tick = self.db._nbtt()
			try:
				e = self[orig]
				e.clear()
				created = False
			except KeyError:
				self.db.query.exist_edge(self.graph.name, orig, self.dest, 0,
											branch, turn, tick, True)
				e = self._make_edge(orig)
			e.update(value)
			self.db._edges_cache.store(self.graph.name, orig, self.dest, 0,
										branch, turn, tick, True)
			self.send(self, key=orig, val=value)

		def __delitem__(self, orig):
			"""Unset the existence of the edge from the given node to mine"""
			branch, turn, tick = self.db._nbtt()
			if 'Multi' in self.graph.__class__.__name__:
				for idx in self[orig]:
					self.db.query.exist_edge(self.graph.name, orig, self.dest,
												idx, branch, turn, tick, False)
					self.db._edges_cache.store(self.graph.name, orig,
												self.dest, idx, branch, turn,
												tick, False)
					self.send(self, key=orig, val=None)
					return
				else:
					raise KeyError("No edges from {}".format(orig))
			self.db.query.exist_edge(self.graph.name, orig, self.dest, 0,
										branch, turn, tick, False)
			self.db._edges_cache.store(self.graph.name, orig, self.dest, 0,
										branch, turn, tick, None)
			self.send(self, key=orig, val=None)


def unwrapped_dict(d):
	return {k: v.unwrap() if hasattr(v, 'unwrap') else v for k, v in d.items()}


class DiGraph(networkx.DiGraph):
	"""A version of the networkx.DiGraph class that stores its state in a
    database.

    """
	adj_cls = DiGraphSuccessorsMapping
	pred_cls = DiGraphPredecessorsMapping
	graph_map_cls = GraphMapping
	node_map_cls = GraphNodeMapping
	_statmap: graph_map_cls
	_nodemap: node_map_cls
	_adjmap: adj_cls
	_predmap: pred_cls

	def __repr__(self):
		return "<{} object named {} containing {} nodes, {} edges>".format(
			self.__class__, self.name, len(self.nodes), len(self.edges))

	def _nodes_state(self):
		return {
			noden: unwrapped_dict(node)
			for noden, node in self._node.items()
		}

	def _edges_state(self):
		ret = {}
		ismul = self.is_multigraph()
		for orig, dests in self.adj.items():
			if orig not in ret:
				ret[orig] = {}
			origd = ret[orig]
			for dest, edge in dests.items():
				if ismul:
					if dest not in origd:
						origd[dest] = edges = {}
					else:
						edges = origd[dest]
					for i, val in edge.items():
						edges[i] = unwrapped_dict(val)
				else:
					origd[dest] = unwrapped_dict(edge)
		return ret

	def _val_state(self):
		return unwrapped_dict(self.graph)

	def __new__(cls, db, name, data=None, **attr):
		if name in db._graph_objs:
			ret = db._graph_objs[name]
			if not isinstance(ret, cls):
				raise EntityCollisionError(
					"Already have a graph named {}, but it's of class {}".
					format(name, type(ret)))
			return ret
		return super(DiGraph, cls).__new__(cls)

	def __init__(self, db, name):  # user shouldn't instantiate directly
		self._name = name
		self.db = db
		if name not in self.db._graph_objs:
			self.db._graph_objs[name] = self

	@property
	def graph(self):
		if not hasattr(self, '_statmap'):
			self._statmap = self.graph_map_cls(self)
		return self._statmap

	@graph.setter
	def graph(self, v):
		self.graph.clear()
		self.graph.update(v)

	@property
	def node(self):
		if not hasattr(self, '_nodemap'):
			self._nodemap = self.node_map_cls(self)
		return self._nodemap

	_node = node

	@property
	def adj(self):
		if not hasattr(self, '_adjmap'):
			self._adjmap = self.adj_cls(self)
		return self._adjmap

	edge = succ = _succ = _adj = adj

	@property
	def pred(self):
		if not hasattr(self, 'pred_cls'):
			raise TypeError("Undirected graph")
		if not hasattr(self, '_predmap'):
			self._predmap = self.pred_cls(self)
		return self._predmap

	_pred = pred

	@property
	def name(self):
		return self._name

	@name.setter
	def name(self, v):
		raise TypeError("graphs can't be renamed")

	def _and_previous(self):
		"""Return a 4-tuple that will usually be (current branch, current
        revision - 1, current branch, current revision), unless
        current revision - 1 is before the start of the current
        branch, in which case the first element will be the parent
        branch.

        """
		branch = self.db.branch
		rev = self.db.rev
		(parent, parent_rev) = self.db.sql('parparrev', branch).fetchone()
		before_branch = parent if parent_rev == rev else branch
		return before_branch, rev - 1, branch, rev

	def remove_node(self, n):
		__doc__ = networkx.DiGraph.remove_node.__doc__
		if n not in self._node:
			raise NetworkXError("The node %s is not in the digraph." % (n, ))
		nbrs = self._succ[n]
		for u in nbrs:
			del self._pred[u][n]  # remove all edges n-u in digraph
		del self._succ[n]  # remove node from succ
		for u in self._pred[n]:
			del self._succ[u][n]  # remove all edges n-u in digraph
		del self._pred[n]  # remove node from pred
		del self._node[n]

	def remove_edge(self, u, v):
		"""Version of remove_edge that's much like normal networkx but only
        deletes once, since the database doesn't keep separate adj and
        succ mappings

        """
		try:
			del self.succ[u][v]
		except KeyError:
			raise NetworkXError("The edge {}-{} is not in the graph.".format(
				u, v))

	def remove_edges_from(self, ebunch):
		"""Version of remove_edges_from that's much like normal networkx but only
        deletes once, since the database doesn't keep separate adj and
        succ mappings

        """
		for e in ebunch:
			(u, v) = e[:2]
			if u in self.succ and v in self.succ[u]:
				del self.succ[u][v]

	def add_edge(self, u, v, attr_dict=None, **attr):
		"""Version of add_edge that only writes to the database once"""
		if attr_dict is None:
			attr_dict = attr
		else:
			try:
				attr_dict.update(attr)
			except AttributeError:
				raise NetworkXError(
					"The attr_dict argument must be a dictionary.")
		if u not in self.node:
			self.node[u] = {}
		if v not in self.node:
			self.node[v] = {}
		if u in self.adj:
			datadict = self.adj[u].get(v, {})
		else:
			self.adj[u] = {v: {}}
			datadict = self.adj[u][v]
		datadict.update(attr_dict)
		self.succ[u][v] = datadict

	def add_edges_from(self, ebunch, attr_dict=None, **attr):
		"""Version of add_edges_from that only writes to the database once"""
		if attr_dict is None:
			attr_dict = attr
		else:
			try:
				attr_dict.update(attr)
			except AttributeError:
				raise NetworkXError("The attr_dict argument must be a dict.")
		for e in ebunch:
			ne = len(e)
			if ne == 3:
				u, v, dd = e
				assert hasattr(dd, "update")
			elif ne == 2:
				u, v = e
				dd = {}
			else:
				raise NetworkXError(
					"Edge tupse {} must be a 2-tuple or 3-tuple.".format(e))
			if u not in self.node:
				self.node[u] = {}
			if v not in self.node:
				self.node[v] = {}
			datadict = self.adj.get(u, {}).get(v, {})
			datadict.update(attr_dict)
			datadict.update(dd)
			self.succ[u][v] = datadict
			assert (u in self.succ)
			assert (v in self.succ[u])

	def clear(self):
		"""Remove all nodes and edges from the graph.

        Unlike the regular networkx implementation, this does *not*
        remove the graph's name. But all the other graph, node, and
        edge attributes go away.

        """
		self.adj.clear()
		self.node.clear()
		self.graph.clear()

	def add_node(self, node_for_adding, **attr):
		if node_for_adding not in self._succ:
			self._succ[node_for_adding] = self.adjlist_inner_dict_factory()
			self._pred[node_for_adding] = self.adjlist_inner_dict_factory()
			self._node[node_for_adding] = self.node_dict_factory()
		self._node[node_for_adding].update(attr)


class GraphsMapping(MutableMapping):

	def __init__(self, orm):
		self.orm = orm

	def __iter__(self):
		return iter(self.orm._graph_objs)

	def __len__(self):
		return len(self.orm._graph_objs)

	def __getitem__(self, item):
		return self.orm.get_graph(item)

	def __setitem__(self, key, value):
		if isinstance(value, networkx.MultiDiGraph):
			self.orm.new_multidigraph(key, data=value)
		elif isinstance(value, networkx.DiGraph):
			self.orm.new_digraph(key, data=value)
		elif isinstance(value, networkx.MultiGraph):
			self.orm.new_multigraph(key, data=value)
		else:
			self.orm.new_graph(key, data=value)

	def __delitem__(self, key):
		self.orm.del_graph(key)



import pytest
import os
from LiSE.allegedb import ORM
import networkx as nx

testgraphs = [nx.chvatal_graph()]
# have to name it after creation because it clears the create_using
path_graph_9 = nx.path_graph(9)
path_graph_9.name = 'path_graph_9'
testgraphs.append(path_graph_9)


@pytest.fixture
def db(tmpdbfile):
	with ORM('sqlite:///' + tmpdbfile) as orm:
		for graph in testgraphs:
			orm.new_digraph(graph.name, graph)
			if not graph.is_directed():
				graph = nx.to_directed(graph)
			assert set(graph.nodes.keys()) == set(
				orm.graph[graph.name].nodes.keys()), \
                               "{}'s nodes changed during instantiation".format(graph.name)
			assert set(graph.edges) == set(orm.graph[graph.name].edges.keys()), \
                               "{}'s edges changed during instantiation".format(graph.name)
	with ORM('sqlite:///' + tmpdbfile) as orm:
		yield orm


def test_basic_load(db):
	for graph in testgraphs:
		if not graph.is_directed():
			graph = nx.to_directed(graph)
		alleged = db.graph[graph.name]
		assert set(graph.nodes.keys()) == set(alleged.nodes.keys(
		)), "{}'s nodes are not the same after load".format(graph.name)
		assert set(graph.edges) == set(
			alleged.edges), "{}'s edges are not the same after load".format(
				graph.name)


def test_keyframe_load(db):
	for graph in testgraphs:
		nodes_kf = db._nodes_cache.keyframe
		assert (graph.name, ) in nodes_kf, "{} not in nodes cache".format(
			graph.name)
		assert 'trunk' in nodes_kf[graph.name,], \
                     "trunk branch not in nodes cache for {}".format(graph.name)
		assert nodes_kf[graph.name,]['trunk'].rev_gettable(0), \
                     "turn 0 not in nodes cache for {}".format(graph.name)
		assert nodes_kf[graph.name,]['trunk'][0].rev_gettable(0), \
                     "tick 0 not in nodes cache for {}".format(graph.name)
		assert db._nodes_cache.keyframe[graph.name, ]['trunk'][0][0] == {
			node: True
			for node in graph.nodes.keys()
		}, "{} not loaded correctly, got {}".format(graph.name,
													nodes_kf['trunk'][0][0])
		edges_kf = db._edges_cache.keyframe
		if graph.is_multigraph():
			for orig in graph.adj:
				for dest in graph.adj[orig]:
					assert (graph.name, orig, dest) in edges_kf, \
                                                   "{} not in edges cache".format(
						(graph.name, orig, dest))
					this_edge = edges_kf[graph.name, orig, dest]
					assert 'trunk' in this_edge, \
                                                   "trunk branch not in edges cache for {}".format(
						(graph.name, orig, dest)
						)
					assert this_edge['trunk'].rev_gettable(0), \
                                                   "turn 0 not in trunk branch of edges cache for {}".format(
						(graph.name, orig, dest)
						)
					assert this_edge['trunk'][0].rev_gettable(0), \
                                                   "tick 0 not in turn 0 of trunk branch of edges cache for {}".format(
						(graph.name, orig, dest)
						)
					assert db._edges_cache.keyframe[graph.name, orig, dest][
						'trunk'][0][0] == {
						idx: True for idx in graph.adj[orig][dest]}, \
                                                   "{} not loaded".format((graph.name, orig, dest))
		else:
			for orig in graph.adj:
				for dest in graph.adj[orig]:
					assert (graph.name, orig, dest) in edges_kf, \
                                                   "{} not in edges cache".format(
						(graph.name, orig, dest)
						)
					this_edge = edges_kf[graph.name, orig, dest]
					assert 'trunk' in this_edge, \
                                                   "trunk branch not in edges cache for {}".format(
						(graph.name, orig, dest)
						)
					assert this_edge['trunk'].rev_gettable(0), \
                                                   "turn 0 not in trunk branch of edges cache for {}".format(
						(graph.name, orig, dest)
						)
					assert this_edge['trunk'][0].rev_gettable(0), \
                                                   "tick 0 not in turn 0 of trunk branch of edges cache for {}".format(
						(graph.name, orig, dest)
						)
					assert db._edges_cache.keyframe[graph.name, orig,
													dest]['trunk'][0][0] == {
														0: True
													}, "{} not loaded".format(
														(graph.name, orig,
															dest))
		for node, vals in graph.nodes.items():
			assert db._node_val_cache.keyframe[graph.name,
												node]['trunk'][0][0] == vals
		for edge in graph.edges:
			if graph.is_multigraph():
				assert db._edge_val_cache.keyframe[
					(graph.name, ) + edge]['trunk'][0][0] == graph.edges[edge]
			else:
				assert db._edge_val_cache.keyframe[(graph.name, ) + edge + (
					0, )]['trunk'][0][0] == graph.edges[edge]


def test_keyframe_unload(tmpdbfile):
	# TODO: test edge cases involving tick-precise unloads
	with ORM('sqlite:///' + tmpdbfile) as orm:
		g = orm.new_digraph('g', nx.grid_2d_graph(3, 3))
		orm.turn = 1
		assert orm._time_is_loaded(*orm._btt())
		assert (
			'g', (0, 0), (0, 1)
			) in orm._edges_cache.keyframe and 0 in \
                           orm._edges_cache.keyframe[
			'g', (0, 0), (0, 1)]['trunk']
		del g.node[1, 1]
		g.add_node('a')
		g.add_edge((0, 0), 'a')
		orm.turn = 2
		orm.snap_keyframe()
		g.add_node((4, 4))
		g.add_edge((3, 3), (4, 4))
		assert (
			'g', (0, 0), (0, 1)
			) in orm._edges_cache.keyframe and 0 in \
                           orm._edges_cache.keyframe[
			'g', (0, 0), (0, 1)]['trunk']
		assert (
			'g',
			) in orm._nodes_cache.keyframe and 'trunk' in \
                           orm._nodes_cache.keyframe[
			'g',] and 0 in orm._nodes_cache.keyframe['g',]['trunk']
		orm.unload()
		assert not orm._time_is_loaded('trunk', 1)
		if 'trunk' in orm._nodes_cache.keyframe['g', ]:
			assert 0 not in orm._nodes_cache.keyframe['g', ]['trunk']
		assert ('g', (0, 0), (0, 1)) in orm._edges_cache.keyframe
		assert 'trunk' in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]
		assert 2 in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]['trunk']
		assert 0 not in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]['trunk']
	with ORM('sqlite:///' + tmpdbfile) as orm:
		assert not orm._time_is_loaded('trunk', 1)
		assert orm._time_is_loaded('trunk', 2, 3)
		assert ('g', (0, 0), (0, 1)) in orm._edges_cache.keyframe
		assert 2 in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]['trunk']
		assert 0 not in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]['trunk']
		g = orm.graph['g']
		if 'trunk' in orm._nodes_cache.keyframe['g', ]:
			assert 0 not in orm._nodes_cache.keyframe['g', ]['trunk']
		if ('g', (0, 0), (0, 1)) in orm._edges_cache.keyframe and \
                      'trunk' in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]:
			assert 0 not in orm._edges_cache.keyframe['g', (0, 0),
														(0, 1)]['trunk']
		assert not orm._time_is_loaded('trunk', 1)
		orm.turn = 0
		assert orm._time_is_loaded('trunk', 1)
		assert 0 in orm._edges_cache.keyframe['g', (0, 0), (0, 1)]['trunk']
		orm.branch = 'u'
		del g.node[1, 2]
		orm.unload()
	with ORM('sqlite:///' + tmpdbfile) as orm:
		assert orm.branch == 'u'
		assert ('g', (1, 1), (1, 2)) not in orm._edges_cache.keyframe \
                           or 'trunk' not in orm._edges_cache.keyframe['g', (1, 1), (1, 2)]
		g = orm.graph['g']
		assert (1, 2) not in g.nodes
		orm.branch = 'trunk'
		assert (1, 2) in g.nodes
		assert ('g', (1, 1), (1, 2)) in orm._edges_cache.keyframe \
                           and 'trunk' in orm._edges_cache.keyframe['g', (1, 1), (1, 2)]



import tempfile
from time import monotonic

from LiSE.handle import EngineHandle
from LiSE.examples.polygons import install

with tempfile.TemporaryDirectory() as dir:
	hand = EngineHandle((dir, ), )
	install(hand._real)
	print(-1)
	hand._real.cache_arrange_queue.join()
	ts = monotonic()
	print(0)
	hand.time_travel('trunk', 2)
	print(1, monotonic() - ts)
	ts = monotonic()
	hand.time_travel('truck', 1)
	print(2, monotonic() - ts)
	ts = monotonic()
	hand.time_travel('trunk', 0)
	print(3, monotonic() - ts)
	ts = monotonic()
	hand.time_travel('truck', 1)
	print(4, monotonic() - ts)
	hand.close()



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The "engine" of LiSE is an object relational mapper with special
stores for game data and entities, as well as properties for manipulating the
flow of time.

"""
from functools import partial
from collections import defaultdict
from types import FunctionType, ModuleType
from typing import Union, Tuple, Any, Set, List, Type, Hashable
from os import PathLike
from abc import ABC, abstractmethod

from networkx import Graph
from blinker import Signal
from .allegedb import ORM as gORM
from .allegedb import StatDictType, NodeValDictType, EdgeValDictType, DeltaType
from .util import sort_set, EntityStatAccessor, AbstractEngine, final_rule
from .xcollections import StringStore, FunctionStore, MethodStore

from . import exc


class InnerStopIteration(StopIteration):
	pass


class DummyEntity(dict):
	"""Something to use in place of a node or edge"""
	__slots__ = ['engine']

	def __init__(self, engine: 'AbstractEngine'):
		super().__init__()
		self.engine = engine


class NextTurn(Signal):
	"""Make time move forward in the simulation.

    Calls ``advance`` repeatedly, returning a list of the rules' return values.

    I am also a ``Signal``, so you can register functions to be
    called when the simulation runs. Pass them to my ``connect``
    method.

    """

	def __init__(self, engine: AbstractEngine):
		super().__init__()
		self.engine = engine

	def __call__(self) -> Tuple[List, DeltaType]:
		engine = self.engine
		start_branch, start_turn, start_tick = engine._btt()
		latest_turn = engine._turns_completed[start_branch]
		if start_turn < latest_turn:
			engine.turn += 1
			self.send(engine,
						branch=engine.branch,
						turn=engine.turn,
						tick=engine.tick)
			return [], engine.get_delta(branch=start_branch,
										turn_from=start_turn,
										turn_to=engine.turn,
										tick_from=start_tick,
										tick_to=engine.tick)
		elif start_turn > latest_turn + 1:
			raise exc.RulesEngineError(
				"Can't run the rules engine on any turn but the latest")
		if start_turn == latest_turn:
			# As a side effect, the following assignment sets the tick
			# to the latest in the new turn, which will be 0 if that
			# turn has not yet been simulated.
			engine.turn += 1
		with engine.advancing():
			for res in iter(engine.advance, final_rule):
				if res:
					engine.universal['last_result'] = res
					engine.universal['last_result_idx'] = 0
					branch, turn, tick = engine._btt()
					self.send(engine, branch=branch, turn=turn, tick=tick)
					return res, engine.get_delta(branch=start_branch,
													turn_from=start_turn,
													turn_to=turn,
													tick_from=start_tick,
													tick_to=tick)
		engine._turns_completed[start_branch] = engine.turn
		if not self.engine.keep_rules_journal:
			engine.query.complete_turn(start_branch, engine.turn)
		if engine.flush_modulus and engine.turn % engine.flush_modulus == 0:
			engine.query.flush()
		if engine.commit_modulus and engine.turn % engine.commit_modulus == 0:
			engine.query.commit()
		self.send(self.engine,
					branch=engine.branch,
					turn=engine.turn,
					tick=engine.tick)
		delta = engine.get_delta(branch=engine.branch,
									turn_from=start_turn,
									turn_to=engine.turn,
									tick_from=start_tick,
									tick_to=engine.tick)
		return [], delta


class AbstractSchema(ABC):
	"""Base class for schemas describing what changes are permitted to the game world"""

	def __init__(self, engine: AbstractEngine):
		self.engine = engine

	@abstractmethod
	def entity_permitted(self, entity):
		raise NotImplementedError

	@abstractmethod
	def stat_permitted(self, turn, entity, key, value):
		raise NotImplementedError


class NullSchema(AbstractSchema):
	"""Schema that permits all changes to the game world"""

	def entity_permitted(self, entity):
		return True

	def stat_permitted(self, turn, entity, key, value):
		return True


class Engine(AbstractEngine, gORM):
	"""LiSE, the Life Simulator Engine.

    Each instance of LiSE maintains a connection to a database
    representing the state of a simulated world. Simulation rules
    within this world are described by lists of Python functions, some
    of which make changes to the world.

    The top-level data structure within LiSE is the character. Most
    data within the world model is kept in some character or other;
    these will quite frequently represent people, but can be readily
    adapted to represent any kind of data that can be comfortably
    described as a graph or a JSON object. Every change to a character
    will be written to the database.

    LiSE tracks history as a series of turns. In each turn, each
    simulation rule is evaluated once for each of the simulated
    entities it's been applied to. World changes in a given turn are
    remembered together, such that the whole world state can be
    rewound: simply set the properties ``branch`` and ``turn`` back to
    what they were just before the change you want to undo.

    Properties:

    - ``branch``: The fork of the timestream that we're on.
    - ``turn``: Units of time that have passed since the sim started.
    - ``time``: ``(branch, turn)``
    - ``tick``: A counter of how many changes have occurred this turn.
    Can be set manually, but is more often set to the last tick in a turn
    as a side effect of setting ``turn``.
    - ``character``: A mapping of :class:`Character` objects by name.
    - ``rule``: A mapping of all rules that have been made.
    - ``rulebook``: A mapping of lists of rules. They are followed in
      their order.  A whole rulebook full of rules may be assigned to
      an entity at once.
    - ``trigger``: Functions that might trigger a rule.
    - ``prereq``: Functions a rule might require to return
      ``True`` for it to run.
    - ``action``: Functions that might manipulate the world
      state as a result of a rule running.
    - ``method``: Extension methods to be added to the engine object.
    - ``function``: Generic functions. All of ``trigger``, ``prereq``,
    ``action``, ``method``, and ``function`` are modules or similar;
    they default to :class:`FunctionStore` objects, which can write
    Python code to the underlying module at runtime.
    - ``string``: A mapping of strings, probably shown to the player
      at some point. Defaults to a :class:`StringStore` object,
      which can alter the underlying JSON file at runtime.
    - ``eternal``: Mapping of arbitrary serializable objects. It isn't
      sensitive to sim-time. A good place to keep game settings.
    - ``universal``: Another mapping of arbitrary serializable
      objects, but this one *is* sensitive to sim-time. Each turn, the
      state of the randomizer is saved here under the key
      ``'rando_state'``.

    """
	from .character import Character
	from .thing import Thing
	from .place import Place
	from .portal import Portal
	from .query import QueryEngine
	char_cls = Character
	thing_cls = Thing
	place_cls = node_cls = Place
	portal_cls = edge_cls = Portal
	query_engine_cls = QueryEngine
	illegal_graph_names = [
		'global', 'eternal', 'universal', 'rulebooks', 'rules'
	]
	illegal_node_names = ['nodes', 'node_val', 'edges', 'edge_val', 'things']

	def __init__(self,
					prefix: Union[PathLike, str] = '.',
					*,
					string: Union[StringStore, dict] = None,
					trigger: Union[FunctionStore, ModuleType] = None,
					prereq: Union[FunctionStore, ModuleType] = None,
					action: Union[FunctionStore, ModuleType] = None,
					function: Union[FunctionStore, ModuleType] = None,
					method: Union[MethodStore, ModuleType] = None,
					connect_string: str = None,
					connect_args: dict = None,
					schema_cls: Type[AbstractSchema] = NullSchema,
					alchemy=False,
					flush_modulus=1,
					commit_modulus=10,
					random_seed: int = None,
					logfun: FunctionType = None,
					clear=False,
					keep_rules_journal=True,
					cache_arranger=True):
		"""Store the connections for the world database and the code database;
        set up listeners; and start a transaction

        :arg prefix: directory containing the simulation and its code;
        defaults to the working directory
        :arg string: module storing strings to be used in the game
        :arg function: module containing utility functions
        :arg method: module containing functions taking this engine as
        first arg
        :arg trigger: module containing trigger functions, taking a LiSE
        entity and returning a boolean for whether to run a rule
        :arg prereq: module containing prereq functions, taking a LiSE entity and
        returning a boolean for whether to permit a rule to run
        :arg action: module containing action functions, taking a LiSE entity and
        mutating it (and possibly the rest of the world)
        :arg connect_string: a rfc1738 URI for a database to connect to;
        if absent, we'll use a SQLite database in the prefix directory.
        With ``alchemy=False`` we can only open SQLite databases,
        in which case ``connect_string`` is just a path to the database--
        unless it's ``":memory:"``, which is an in-memory database that
        won't be saved
        :arg connect_args: dictionary of keyword arguments for the
        database connection
        :arg schema: a Schema class that determines which changes to allow to
        the world; used when a player should not be able to change just anything.
        Defaults to `NullSchema`
        :arg alchemy: whether to use SQLAlchemy to connect to the
        database. If False, LiSE can only use SQLite
        :arg flush_modulus: LiSE will put pending changes into the database
        transaction every ``flush_modulus`` turns, default 1. If `None`,
        only flush on commit
        :arg commit_modulus: LiSE will commit changes to disk every
        ``commit_modulus`` turns, default 10. If `None`, only commit
        on close or manual call to `commit`
        :arg random_seed: a number to initialize the randomizer
        :arg logfun: an optional function taking arguments
        ``level, message``, which should log `message` somehow
        :arg clear: whether to delete *any and all* existing data
        and code in ``prefix``. Use with caution!
        :arg keep_rules_journal: Boolean; if true (default), keep
        information on the behavior of the rules engine in the database.
        Makes the database rather large, but useful for debugging.
        :arg cache_arranger: If true (default), start a background
        process that indexes the caches to make time travel faster
        when it's to points we anticipate. If you use this, you can
        specify some other point in time to index by putting the
        `(branch, turn, tick)` in my `cache_arrange_queue`.

        """
		if logfun is None:
			from logging import getLogger
			logger = getLogger("Life Sim Engine")

			def logfun(level, msg):
				getattr(logger, level)(msg)

		self.log = logfun
		import os
		from .xcollections import StringStore
		if connect_args is None:
			connect_args = {}
		if not os.path.exists(prefix):
			os.mkdir(prefix)
		if not os.path.isdir(prefix):
			raise FileExistsError("Need a directory")
		self.keep_rules_journal = keep_rules_journal
		if string:
			self.string = string
		else:
			self._string_file = os.path.join(prefix, 'strings.json')
			if clear and os.path.exists(self._string_file):
				os.remove(self._string_file)
		if function:
			self.function = function
		else:
			self._function_file = os.path.join(prefix, 'function.py')
			if clear and os.path.exists(self._function_file):
				os.remove(self._function_file)
		if method:
			self.method = method
		else:
			self._method_file = os.path.join(prefix, 'method.py')
			if clear and os.path.exists(self._method_file):
				os.remove(self._method_file)
		if trigger:
			self.trigger = trigger
		else:
			self._trigger_file = os.path.join(prefix, 'trigger.py')
			if clear and os.path.exists(self._trigger_file):
				os.remove(self._trigger_file)
		if prereq:
			self.prereq = prereq
		else:
			self._prereq_file = os.path.join(prefix, 'prereq.py')
			if clear and os.path.exists(self._prereq_file):
				os.remove(self._prereq_file)
		if action:
			self.action = action
		else:
			self._action_file = os.path.join(prefix, 'action.py')
			if clear and os.path.exists(self._action_file):
				os.remove(self._action_file)
		self.schema = schema_cls(self)
		if connect_string and not alchemy:
			connect_string = connect_string.split('sqlite:///')[-1]
		super().__init__(connect_string or os.path.join(prefix, 'world.db'),
							sqlfilename=os.path.join(
								os.path.dirname(os.path.abspath(__file__)),
								'sqlite.json'),
							clear=clear,
							connect_args=connect_args,
							alchemy=alchemy,
							cache_arranger=cache_arranger)
		self._things_cache.setdb = self.query.set_thing_loc
		self._universal_cache.setdb = self.query.universal_set
		self._rulebooks_cache.setdb = self.query.rulebook_set
		self.eternal = self.query.globl
		if hasattr(self, '_string_file'):
			self.string = StringStore(
				self.query, self._string_file,
				self.eternal.setdefault('language', 'eng'))
		self.next_turn = NextTurn(self)
		self.commit_modulus = commit_modulus
		self.flush_modulus = flush_modulus
		self.random_seed = random_seed
		self._rules_iter = self._follow_rules()
		# set up the randomizer
		from random import Random
		self._rando = Random()
		if 'rando_state' in self.universal:
			self._rando.setstate(self.universal['rando_state'])
		else:
			self._rando.seed(self.random_seed)
			self.universal['rando_state'] = self._rando.getstate()
		if hasattr(self.method, 'init'):
			self.method.init(self)
		if cache_arranger:
			self._start_cache_arranger()

	def _start_cache_arranger(self) -> None:
		for branch, (parent, turn_start, tick_start, turn_end,
						tick_end) in self._branches.items():
			self.cache_arrange_queue.put((branch, turn_start, tick_start))
			if (turn_start, tick_start) != (turn_end, tick_end):
				self.cache_arrange_queue.put((branch, turn_end, tick_end))
		if not self._cache_arrange_thread.is_alive():
			self._cache_arrange_thread.start()

	def _init_load(self) -> None:
		from .rule import Rule
		q = self.query
		super()._init_load()
		self._unitness_cache.load(q.units_dump())
		self._universal_cache.load(q.universals_dump())
		self._rulebooks_cache.load(q.rulebooks_dump())
		self._characters_rulebooks_cache.load(q.character_rulebook_dump())
		self._units_rulebooks_cache.load(q.unit_rulebook_dump())
		self._characters_things_rulebooks_cache.load(
			q.character_thing_rulebook_dump())
		self._characters_places_rulebooks_cache.load(
			q.character_place_rulebook_dump())
		self._characters_portals_rulebooks_cache.load(
			q.character_portal_rulebook_dump())
		self._nodes_rulebooks_cache.load(q.node_rulebook_dump())
		self._portals_rulebooks_cache.load(q.portal_rulebook_dump())
		self._triggers_cache.load(q.rule_triggers_dump())
		self._prereqs_cache.load(q.rule_prereqs_dump())
		self._actions_cache.load(q.rule_actions_dump())
		store_crh = self._character_rules_handled_cache.store
		for row in q.character_rules_handled_dump():
			store_crh(*row, loading=True)
		store_arh = self._unit_rules_handled_cache.store
		for row in q.unit_rules_handled_dump():
			store_arh(*row, loading=True)
		store_ctrh = self._character_thing_rules_handled_cache.store
		for row in q.character_thing_rules_handled_dump():
			store_ctrh(*row, loading=True)
		store_cprh = self._character_place_rules_handled_cache.store
		for row in q.character_place_rules_handled_dump():
			store_cprh(*row, loading=True)
		store_cporh = self._character_portal_rules_handled_cache.store
		for row in q.character_portal_rules_handled_dump():
			store_cporh(*row, loading=True)
		store_cnrh = self._node_rules_handled_cache.store
		for row in q.node_rules_handled_dump():
			store_cnrh(*row, loading=True)
		store_porh = self._portal_rules_handled_cache.store
		for row in q.portal_rules_handled_dump():
			store_porh(*row, loading=True)
		self._turns_completed.update(q.turns_completed_dump())
		self._rules_cache = {
			name: Rule(self, name, create=False)
			for name in q.rules_dump()
		}

	def _load_at(self, branch: str, turn: int, tick: int) -> None:
		latest_past_keyframe, earliest_future_keyframe, keyframed, noderows, edgerows, graphvalrows, nodevalrows, edgevalrows = super(
		)._load_at(branch, turn, tick)
		thingrows = []

		def build_thingrows(graf, windows):
			if not windows:
				return
			if len(windows) == 1:
				btt = windows[0]
				thingrows.extend(load_things(graf, *btt))
				return
			for window in reversed(windows):
				thingrows.extend(load_things(graf, *window))

		load_things = self.query.load_things
		if latest_past_keyframe is None:
			# Load thing data from the beginning of time to now
			for graph in self.graph:
				build_thingrows(
					graph,
					self._build_loading_windows('trunk', 0, 0, branch, turn,
												tick))
		else:
			past_branch, past_turn, past_tick = latest_past_keyframe
			if earliest_future_keyframe is None:
				# Load thing data from the keyframe to now
				for graph in self.graph:
					build_thingrows(
						graph,
						self._build_loading_windows(past_branch, past_turn,
													past_tick, branch, turn,
													tick))
			else:
				# Load thing data between the two keyframes
				future_branch, future_turn, future_tick = earliest_future_keyframe

				for graph in self.graph:
					build_thingrows(
						graph,
						self._build_loading_windows(past_branch, past_turn,
													past_tick, future_branch,
													future_turn, future_tick))
		if thingrows:
			with self.batch():
				self._things_cache.load(thingrows)
		else:
			self.warning(f"No thing data at {branch, turn, tick}")

	def _init_caches(self) -> None:
		from .xcollections import (FunctionStore, CharacterMapping,
									UniversalMapping)
		from .cache import (
			NodeContentsCache, InitializedCache, EntitylessCache,
			InitializedEntitylessCache, UnitnessCache, UnitRulesHandledCache,
			CharacterThingRulesHandledCache, CharacterPlaceRulesHandledCache,
			CharacterPortalRulesHandledCache, NodeRulesHandledCache,
			PortalRulesHandledCache, CharacterRulesHandledCache, ThingsCache)
		from .rule import AllRuleBooks, AllRules

		super()._init_caches()
		self._things_cache = ThingsCache(self)
		self._node_contents_cache = NodeContentsCache(self)
		self.character = self.graph = CharacterMapping(self)
		self._universal_cache = EntitylessCache(self)
		self._universal_cache.name = 'universal_cache'
		self._rulebooks_cache = InitializedEntitylessCache(self)
		self._rulebooks_cache.name = 'rulebooks_cache'
		self._characters_rulebooks_cache = InitializedEntitylessCache(self)
		self._characters_rulebooks_cache.name = 'characters_rulebooks_cache'
		self._units_rulebooks_cache = InitializedEntitylessCache(self)
		self._units_rulebooks_cache.name = 'units_rulebooks_cache'
		self._characters_things_rulebooks_cache = InitializedEntitylessCache(
			self)
		self._characters_things_rulebooks_cache.name = 'characters_things_rulebooks_cache'
		self._characters_places_rulebooks_cache = InitializedEntitylessCache(
			self)
		self._characters_places_rulebooks_cache.name = 'characters_places_rulebooks_cache'
		self._characters_portals_rulebooks_cache = InitializedEntitylessCache(
			self)
		self._characters_portals_rulebooks_cache.name = 'characters_portals_rulebooks_cache'
		self._nodes_rulebooks_cache = InitializedCache(self)
		self._nodes_rulebooks_cache.name = 'nodes_rulebooks_cache'
		self._portals_rulebooks_cache = InitializedCache(self)
		self._portals_rulebooks_cache.name = 'portals_rulebooks_cache'
		self._triggers_cache = InitializedEntitylessCache(self)
		self._triggers_cache.name = 'triggers_cache'
		self._prereqs_cache = InitializedEntitylessCache(self)
		self._prereqs_cache.name = 'prereqs_cache'
		self._actions_cache = InitializedEntitylessCache(self)
		self._actions_cache.name = 'actions_cache'
		self._node_rules_handled_cache = NodeRulesHandledCache(self)
		self._node_rules_handled_cache.name = 'node_rules_handled_cache'
		self._portal_rules_handled_cache = PortalRulesHandledCache(self)
		self._portal_rules_handled_cache.name = 'portal_rules_handled_cache'
		self._character_rules_handled_cache = CharacterRulesHandledCache(self)
		self._character_rules_handled_cache.name = 'character_rules_handled_cache'
		self._unit_rules_handled_cache = UnitRulesHandledCache(self)
		self._unit_rules_handled_cache.name = 'unit_rules_handled_cache'
		self._character_thing_rules_handled_cache = CharacterThingRulesHandledCache(
			self)
		self._character_thing_rules_handled_cache.name = 'character_thing_rules_handled_cache'
		self._character_place_rules_handled_cache = CharacterPlaceRulesHandledCache(
			self)
		self._character_place_rules_handled_cache.name = 'character_place_rules_handled_cache'
		self._character_portal_rules_handled_cache = CharacterPortalRulesHandledCache(
			self)
		self._character_portal_rules_handled_cache.name = 'character_portal_rules_handled_cache'
		self._unitness_cache = UnitnessCache(self)
		self._unitness_cache.name = 'unitness_cache'
		self._turns_completed = defaultdict(lambda: max((0, self.turn - 1)))
		self._turns_completed_previous = self._turns_completed.copy()
		"""The last turn when the rules engine ran in each branch"""
		self.universal = UniversalMapping(self)
		if hasattr(self, '_action_file'):
			self.action = FunctionStore(self._action_file)
		if hasattr(self, '_prereq_file'):
			self.prereq = FunctionStore(self._prereq_file)
		if hasattr(self, '_trigger_file'):
			self.trigger = FunctionStore(self._trigger_file)
		if hasattr(self, '_function_file'):
			self.function = FunctionStore(self._function_file)
		if hasattr(self, '_method_file'):
			self.method = FunctionStore(self._method_file)
		self.rule = AllRules(self)
		self.rulebook = AllRuleBooks(self)

	def _load_graphs(self) -> None:
		for charn in self.query.characters():
			self._graph_objs[charn] = self.char_cls(self,
													charn,
													init_rulebooks=False)

	def _make_node(self, graph: Character,
					node: Hashable) -> Union[thing_cls, place_cls]:
		if self._is_thing(graph.name, node):
			return self.thing_cls(graph, node)
		else:
			return self.place_cls(graph, node)

	def _make_edge(self,
					graph: Character,
					orig: Union[thing_cls, place_cls],
					dest: Union[thing_cls, place_cls],
					idx=0) -> portal_cls:
		return self.portal_cls(graph, orig, dest)

	def get_delta(self, branch: str, turn_from: int, tick_from: int,
					turn_to: int, tick_to: int) -> DeltaType:
		"""Get a dictionary describing changes to the world.

        Most keys will be character names, and their values will be
        dictionaries of the character's stats' new values, with ``None``
        for deleted keys. Characters' dictionaries have special keys
        'nodes' and 'edges' which contain booleans indicating whether
        the node or edge exists at the moment, and 'node_val' and
        'edge_val' for the stats of those entities. For edges (also
        called portals) these dictionaries are two layers deep, keyed
        first by the origin, then by the destination.

        Characters also have special keys for the various rulebooks
        they have:

        * 'character_rulebook'
        * 'unit_rulebook'
        * 'character_thing_rulebook'
        * 'character_place_rulebook'
        * 'character_portal_rulebook'

        And each node and edge may have a 'rulebook' stat of its own.
        If a node is a thing, it gets a 'location'; when the 'location'
        is deleted, that means it's back to being a place.

        Keys at the top level that are not character names:

        * 'rulebooks', a dictionary keyed by the name of each changed
        rulebook, the value being a list of rule names
        * 'rules', a dictionary keyed by the name of each changed rule,
        containing any of the lists 'triggers', 'prereqs', and 'actions'

        """
		from .allegedb.window import update_window, update_backward_window
		if not isinstance(branch, str):
			raise TypeError("branch must be str")
		for arg in (turn_from, tick_from, turn_to, tick_to):
			if not isinstance(arg, int):
				raise TypeError("turn and tick must be int")
		if turn_from == turn_to:
			return self.get_turn_delta(branch,
										turn_to,
										tick_to,
										start_tick=tick_from)
		delta = super().get_delta(branch, turn_from, tick_from, turn_to,
									tick_to)
		if turn_from < turn_to:
			updater = partial(update_window, turn_from, tick_from, turn_to,
								tick_to)
			attribute = 'settings'
			tick_to += 1
		else:
			updater = partial(update_backward_window, turn_from, tick_from,
								turn_to, tick_to)
			attribute = 'presettings'
		univbranches = getattr(self._universal_cache, attribute)
		avbranches = getattr(self._unitness_cache, attribute)
		thbranches = getattr(self._things_cache, attribute)
		rbbranches = getattr(self._rulebooks_cache, attribute)
		trigbranches = getattr(self._triggers_cache, attribute)
		preqbranches = getattr(self._prereqs_cache, attribute)
		actbranches = getattr(self._actions_cache, attribute)
		charrbbranches = getattr(self._characters_rulebooks_cache, attribute)
		avrbbranches = getattr(self._units_rulebooks_cache, attribute)
		charthrbbranches = getattr(self._characters_things_rulebooks_cache,
									attribute)
		charplrbbranches = getattr(self._characters_places_rulebooks_cache,
									attribute)
		charporbbranches = getattr(self._characters_portals_rulebooks_cache,
									attribute)
		noderbbranches = getattr(self._nodes_rulebooks_cache, attribute)
		edgerbbranches = getattr(self._portals_rulebooks_cache, attribute)

		def upduniv(_, key, val):
			delta.setdefault('universal', {})[key] = val

		if branch in univbranches:
			updater(upduniv, univbranches[branch])

		def updav(char, graph, node, av):
			delta.setdefault(char, {}).setdefault('units', {}).setdefault(
				graph, {})[node] = bool(av)

		if branch in avbranches:
			updater(updav, avbranches[branch])

		def updthing(char, thing, loc):
			if (char in delta and 'nodes' in delta[char]
				and thing in delta[char]['nodes']
				and not delta[char]['nodes'][thing]):
				return
			thingd = delta.setdefault(char, {}).setdefault('node_val',
															{}).setdefault(
																thing, {})
			thingd['location'] = loc

		if branch in thbranches:
			updater(updthing, thbranches[branch])

		def updrb(_, rulebook, rules):
			delta.setdefault('rulebooks', {})[rulebook] = rules

		if branch in rbbranches:
			updater(updrb, rbbranches[branch])

		def updru(key, _, rule, funs):
			delta.setdefault('rules', {}).setdefault(rule, {})[key] = funs

		if branch in trigbranches:
			updater(partial(updru, 'triggers'), trigbranches[branch])

		if branch in preqbranches:
			updater(partial(updru, 'prereqs'), preqbranches[branch])

		if branch in actbranches:
			updater(partial(updru, 'actions'), actbranches[branch])

		def updcrb(key, _, character, rulebook):
			delta.setdefault(character, {})[key] = rulebook

		if branch in charrbbranches:
			updater(partial(updcrb, 'character_rulebook'),
					charrbbranches[branch])

		if branch in avrbbranches:
			updater(partial(updcrb, 'unit_rulebook'), avrbbranches[branch])

		if branch in charthrbbranches:
			updater(partial(updcrb, 'character_thing_rulebook'),
					charthrbbranches[branch])

		if branch in charplrbbranches:
			updater(partial(updcrb, 'character_place_rulebook'),
					charplrbbranches[branch])

		if branch in charporbbranches:
			updater(partial(updcrb, 'character_portal_rulebook'),
					charporbbranches[branch])

		def updnoderb(character, node, rulebook):
			if (character in delta and 'nodes' in delta[character]
				and node in delta[character]['nodes']
				and not delta[character]['nodes'][node]):
				return
			delta.setdefault(character,
								{}).setdefault('node_val', {}).setdefault(
									node, {})['rulebook'] = rulebook

		if branch in noderbbranches:
			updater(updnoderb, noderbbranches[branch])

		def updedgerb(character, orig, dest, rulebook):
			if (character in delta and 'edges' in delta[character]
				and orig in delta[character]['edges']
				and dest in delta[character]['edges'][orig]
				and not delta[character]['edges'][orig][dest]):
				return
			delta.setdefault(character,
								{}).setdefault('edge_val', {}).setdefault(
									orig,
									{}).setdefault(dest,
													{})['rulebook'] = rulebook

		if branch in edgerbbranches:
			updater(updedgerb, edgerbbranches[branch])

		return delta

	def get_turn_delta(self,
						branch: str = None,
						turn: int = None,
						tick: int = None,
						start_tick=0) -> DeltaType:
		"""Get a dictionary of changes to the world within a given turn

        Defaults to the present turn, and stops at the present tick
        unless specified.

        See the documentation for ``get_delta`` for a detailed
        description of the delta format.

        :arg branch: branch of history, defaulting to the present branch
        :arg turn: turn within the branch, defaulting to the present
        turn
        :arg tick: tick at which to stop the delta, defaulting to the
        present tick
        :arg start_tick: tick at which to start the delta, default 0

        """
		branch = branch or self.branch
		turn = turn or self.turn
		tick = tick or self.tick
		if tick == start_tick:
			return {}
		delta = super().get_turn_delta(branch, turn, start_tick, tick)
		if start_tick < tick:
			attribute = 'settings'
			tick += 1
		else:
			attribute = 'presettings'
		avatarness_settings = getattr(self._unitness_cache, attribute)
		things_settings = getattr(self._things_cache, attribute)
		rulebooks_settings = getattr(self._rulebooks_cache, attribute)
		triggers_settings = getattr(self._triggers_cache, attribute)
		prereqs_settings = getattr(self._prereqs_cache, attribute)
		actions_settings = getattr(self._actions_cache, attribute)
		character_rulebooks_settings = getattr(
			self._characters_rulebooks_cache, attribute)
		avatar_rulebooks_settings = getattr(self._units_rulebooks_cache,
											attribute)
		character_thing_rulebooks_settings = getattr(
			self._characters_things_rulebooks_cache, attribute)
		character_place_rulebooks_settings = getattr(
			self._characters_places_rulebooks_cache, attribute)
		character_portal_rulebooks_settings = getattr(
			self._characters_portals_rulebooks_cache, attribute)
		node_rulebooks_settings = getattr(self._nodes_rulebooks_cache,
											attribute)
		portal_rulebooks_settings = getattr(self._portals_rulebooks_cache,
											attribute)
		if branch in avatarness_settings and turn in avatarness_settings[
			branch]:
			for chara, graph, node, is_av in avatarness_settings[branch][turn][
				start_tick:tick]:
				delta.setdefault(chara, {}).setdefault('units', {}).setdefault(
					graph, {})[node] = is_av
		if branch in things_settings and turn in things_settings[branch]:
			for chara, thing, location in things_settings[branch][turn][
				start_tick:tick]:
				thingd = delta.setdefault(chara, {}).setdefault(
					'node_val', {}).setdefault(thing, {})
				thingd['location'] = location
		delta['rulebooks'] = rbdif = {}
		if branch in rulebooks_settings and turn in rulebooks_settings[branch]:
			for _, rulebook, rules in rulebooks_settings[branch][turn][
				start_tick:tick]:
				rbdif[rulebook] = rules
		delta['rules'] = rdif = {}
		if branch in triggers_settings and turn in triggers_settings[branch]:
			for _, rule, funs in triggers_settings[branch][turn][
				start_tick:tick]:
				rdif.setdefault(rule, {})['triggers'] = funs
		if branch in prereqs_settings and turn in prereqs_settings[branch]:
			for _, rule, funs in prereqs_settings[branch][turn][
				start_tick:tick]:
				rdif.setdefault(rule, {})['prereqs'] = funs
		if branch in actions_settings and turn in actions_settings[branch]:
			for _, rule, funs in actions_settings[branch][turn][
				start_tick:tick]:
				rdif.setdefault(rule, {})['actions'] = funs

		if branch in character_rulebooks_settings and turn in character_rulebooks_settings[
			branch]:
			for _, character, rulebook in character_rulebooks_settings[branch][
				turn][start_tick:tick]:
				delta.setdefault(character,
									{})['character_rulebook'] = rulebook
		if branch in avatar_rulebooks_settings and turn in avatar_rulebooks_settings[
			branch]:
			for _, character, rulebook in avatar_rulebooks_settings[branch][
				turn][start_tick:tick]:
				delta.setdefault(character, {})['unit_rulebook'] = rulebook
		if branch in character_thing_rulebooks_settings and turn in character_thing_rulebooks_settings[
			branch]:
			for _, character, rulebook in character_thing_rulebooks_settings[
				branch][turn][start_tick:tick]:
				delta.setdefault(character,
									{})['character_thing_rulebook'] = rulebook
		if branch in character_place_rulebooks_settings and turn in character_place_rulebooks_settings[
			branch]:
			for _, character, rulebook in character_place_rulebooks_settings[
				branch][turn][start_tick:tick]:
				delta.setdefault(character,
									{})['character_place_rulebook'] = rulebook
		if branch in character_portal_rulebooks_settings and turn in character_portal_rulebooks_settings[
			branch]:
			for _, character, rulebook in character_portal_rulebooks_settings[
				branch][turn][start_tick:tick]:
				delta.setdefault(character,
									{})['character_portal_rulebook'] = rulebook

		if branch in node_rulebooks_settings and turn in node_rulebooks_settings[
			branch]:
			for character, node, rulebook in node_rulebooks_settings[branch][
				turn][start_tick:tick]:
				delta.setdefault(character,
									{}).setdefault('node_val', {}).setdefault(
										node, {})['rulebook'] = rulebook
		if branch in portal_rulebooks_settings and turn in portal_rulebooks_settings[
			branch]:
			for character, orig, dest, rulebook in portal_rulebooks_settings[
				branch][turn][start_tick:tick]:
				delta.setdefault(character, {}).setdefault(
					'edge_val',
					{}).setdefault(orig,
									{}).setdefault(dest,
													{})['rulebook'] = rulebook
		return delta

	def _del_rulebook(self, rulebook):
		raise NotImplementedError("Can't delete rulebooks yet")

	def _remember_unitness(self,
							character: Character,
							graph: Character,
							node: Union[thing_cls, place_cls],
							is_unit=True,
							branch: str = None,
							turn: int = None,
							tick: int = None) -> None:
		"""Use this to record a change in unitness.

        Should be called whenever a node that wasn't an unit of a
        character now is, and whenever a node that was an unit of a
        character now isn't.

        ``character`` is the one using the node as an unit,
        ``graph`` is the character the node is in.

        """
		branch = branch or self.branch
		turn = turn or self.turn
		tick = tick or self.tick
		self._unitness_cache.store(character, graph, node, branch, turn, tick,
									is_unit)
		self.query.unit_set(character, graph, node, branch, turn, tick,
							is_unit)

	@property
	def stores(self):
		return (self.action, self.prereq, self.trigger, self.function,
				self.method, self.string)

	def debug(self, msg: str) -> None:
		"""Log a message at level 'debug'"""
		self.log('debug', msg)

	def info(self, msg: str) -> None:
		"""Log a message at level 'info'"""
		self.log('info', msg)

	def warning(self, msg: str) -> None:
		"""Log a message at level 'warning'"""
		self.log('warning', msg)

	def error(self, msg: str) -> None:
		"""Log a message at level 'error'"""
		self.log('error', msg)

	def critical(self, msg: str) -> None:
		"""Log a message at level 'critical'"""
		self.log('critical', msg)

	def flush(self) -> None:
		self.query.flush()

	def commit(self) -> None:
		try:
			self.universal['rando_state'] = self._rando.getstate()
		except exc.OutOfTimelineError:
			branch, turn, tick = self.branch, self.turn, self.tick
			self.turn = self._branches[branch][3]
			self.universal['rando_state'] = self._rando.getstate()
			self.turn = turn
			self.tick = tick
		turns_completed_previous = self._turns_completed_previous
		turns_completed = self._turns_completed
		set_turn_completed = self.query.set_turn_completed
		for branch, turn_late in turns_completed.items():
			turn_early = turns_completed_previous.get(branch)
			if turn_late != turn_early:
				assert turn_early is None or turn_late > turn_early, "Incoherent turns_completed cache"
				set_turn_completed(branch, turn_late)
		self._turns_completed_previous = turns_completed.copy()
		super().commit()

	def close(self) -> None:
		"""Commit changes and close the database."""
		import sys
		import os
		for store in self.stores:
			if hasattr(store, 'save'):
				store.save(reimport=False)
			path, filename = os.path.split(store._filename)
			modname = filename[:-3]
			if modname in sys.modules:
				del sys.modules[modname]
		super().close()

	def __enter__(self):
		"""Return myself. For compatibility with ``with`` semantics."""
		return self

	def __exit__(self, *args):
		"""Close on exit."""
		self.close()

	def _set_branch(self, v: str) -> None:
		oldrando = self.universal.get('rando_state')
		super()._set_branch(v)
		newrando = self.universal.get('rando_state')
		if newrando and newrando != oldrando:
			self._rando.setstate(newrando)
		self.time.send(self.time, branch=self._obranch, turn=self._oturn)

	def _set_turn(self, v: int) -> None:
		turn_end = self._branch_end_plan[self.branch]
		if v > turn_end + 1:
			raise exc.OutOfTimelineError(
				f"The turn {v} is after the end of the branch {self.branch}. "
				f"Go to turn {turn_end + 1} and simulate with `next_turn`.",
				self.branch, self.turn, self.tick, self.branch, v, self.tick)
		oldrando = self.universal.get('rando_state')
		oldturn = self._oturn
		super()._set_turn(v)
		newrando = self.universal.get('rando_state')
		if v > oldturn and newrando and newrando != oldrando:
			self._rando.setstate(newrando)
		self.time.send(self.time, branch=self._obranch, turn=self._oturn)

	def _set_tick(self, v: int) -> None:
		tick_end = self._turn_end_plan[self.branch, self.turn]
		if v > tick_end + 1:
			raise exc.OutOfTimelineError(
				f"The tick {v} is after the end of the turn {self.turn}. "
				f"Go to tick {tick_end + 1} and simulate with `next_turn`.",
				self.branch, self.turn, self.tick, self.branch, self.turn, v)
		oldrando = self.universal.get('rando_state')
		oldtick = self._otick
		super()._set_tick(v)
		newrando = self.universal.get('rando_state')
		if v > oldtick and newrando and newrando != oldrando:
			self._rando.setstate(newrando)

	def _handled_char(self, charn: Hashable, rulebook: Hashable,
						rulen: Hashable, branch: str, turn: int,
						tick: int) -> None:
		try:
			self._character_rules_handled_cache.store(charn, rulebook, rulen,
														branch, turn, tick)
		except ValueError:
			assert rulen in self._character_rules_handled_cache.handled[
				charn, rulebook, branch, turn]
			return
		self.query.handled_character_rule(charn, rulebook, rulen, branch, turn,
											tick)

	def _handled_av(self, character: Hashable, graph: Hashable,
					avatar: Hashable, rulebook: Hashable, rule: Hashable,
					branch: str, turn: int, tick: int) -> None:
		try:
			self._unit_rules_handled_cache.store(character, graph, avatar,
													rulebook, rule, branch,
													turn, tick)
		except ValueError:
			assert rule in self._unit_rules_handled_cache.handled[character,
																	graph,
																	avatar,
																	rulebook,
																	branch,
																	turn]
			return
		self.query.handled_unit_rule(character, rulebook, rule, graph, avatar,
										branch, turn, tick)

	def _handled_char_thing(self, character: Hashable, thing: Hashable,
							rulebook: Hashable, rule: Hashable, branch: str,
							turn: int, tick: int) -> None:
		try:
			self._character_thing_rules_handled_cache.store(
				character, thing, rulebook, rule, branch, turn, tick)
		except ValueError:
			assert rule in self._character_thing_rules_handled_cache.handled[
				character, thing, rulebook, branch, turn]
			return
		self.query.handled_character_thing_rule(character, rulebook, rule,
												thing, branch, turn, tick)

	def _handled_char_place(self, character: Hashable, place: Hashable,
							rulebook: Hashable, rule: Hashable, branch: str,
							turn: int, tick: int) -> None:
		try:
			self._character_place_rules_handled_cache.store(
				character, place, rulebook, rule, branch, turn, tick)
		except ValueError:
			assert rule in self._character_place_rules_handled_cache.handled[
				character, place, rulebook, branch, turn]
			return
		self.query.handled_character_place_rule(character, rulebook, rule,
												place, branch, turn, tick)

	def _handled_char_port(self, character: Hashable, orig: Hashable,
							dest: Hashable, rulebook: Hashable, rule: Hashable,
							branch: str, turn: int, tick: int) -> None:
		try:
			self._character_portal_rules_handled_cache.store(
				character, orig, dest, rulebook, rule, branch, turn, tick)
		except ValueError:
			assert rule in self._character_portal_rules_handled_cache.handled[
				character, orig, dest, rulebook, branch, turn]
			return
		self.query.handled_character_portal_rule(character, orig, dest,
													rulebook, rule, branch,
													turn, tick)

	def _handled_node(self, character: Hashable, node: Hashable,
						rulebook: Hashable, rule: Hashable, branch: str,
						turn: int, tick: int) -> None:
		try:
			self._node_rules_handled_cache.store(character, node, rulebook,
													rule, branch, turn, tick)
		except ValueError:
			assert rule in self._node_rules_handled_cache.handled[character,
																	node,
																	rulebook,
																	branch,
																	turn]
			return
		self.query.handled_node_rule(character, node, rulebook, rule, branch,
										turn, tick)

	def _handled_portal(self, character: Hashable, orig: Hashable,
						dest: Hashable, rulebook: Hashable, rule: Hashable,
						branch: str, turn: int, tick: int) -> None:
		try:
			self._portal_rules_handled_cache.store(character, orig, dest,
													rulebook, rule, branch,
													turn, tick)
		except ValueError:
			assert rule in self._portal_rules_handled_cache.handled[character,
																	orig, dest,
																	rulebook,
																	branch,
																	turn]
			return
		self.query.handled_portal_rule(character, orig, dest, rulebook, rule,
										branch, turn, tick)

	def _follow_rules(self):
		# TODO: roll back changes done by rules that raise an exception
		# TODO: if there's a paradox while following some rule,
		#  start a new branch, copying handled rules
		from collections import defaultdict
		branch, turn, tick = self._btt()
		charmap = self.character
		rulemap = self.rule
		todo = defaultdict(list)

		def check_triggers(rule, handled_fun, entity):
			for trigger in rule.triggers:
				res = trigger(entity)
				if res:
					return True
			else:
				handled_fun(self.tick)
				return False

		def check_prereqs(rule, handled_fun, entity):
			if not entity:
				return False
			for prereq in rule.prereqs:
				res = prereq(entity)
				if not res:
					handled_fun(self.tick)
					return False
			return True

		def do_actions(rule, handled_fun, entity):
			actres = []
			for action in rule.actions:
				res = action(entity)
				if res:
					actres.append(res)
				if not entity:
					break
			handled_fun(self.tick)
			return actres

		# TODO: triggers that don't mutate anything should be
		#  evaluated in parallel
		for (charactername, rulebook, rulename
				) in self._character_rules_handled_cache.iter_unhandled_rules(
					branch, turn, tick):
			if charactername not in charmap:
				continue
			self.debug(
				f"checking triggers for character rule: {charactername}, {rulebook}, {rulename}"
			)
			rule = rulemap[rulename]
			handled = partial(self._handled_char, charactername, rulebook,
								rulename, branch, turn)
			entity = charmap[charactername]
			if check_triggers(rule, handled, entity):
				self.debug(
					f"character rule triggered: {charactername}, {rulebook}, {rulename}"
				)
				todo[rulebook].append((rule, handled, entity))
		avcache_retr = self._unitness_cache._base_retrieve
		node_exists = self._node_exists
		get_node = self._get_node
		for (charn, graphn, avn, rulebook,
				rulen) in self._unit_rules_handled_cache.iter_unhandled_rules(
					branch, turn, tick):
			if not node_exists(graphn, avn) or avcache_retr(
				(charn, graphn, avn, branch, turn, tick)) in (KeyError, None):
				continue
			self.debug(
				f"checking triggers for unit rule: {charn, graphn, avn, rulebook, rulen}"
			)
			rule = rulemap[rulen]
			handled = partial(self._handled_av, charn, graphn, avn, rulebook,
								rulen, branch, turn)
			entity = get_node(graphn, avn)
			if check_triggers(rule, handled, entity):
				self.debug(
					f"unit rule triggered: {charn, graphn, avn, rulebook, rulen}"
				)
				todo[rulebook].append((rule, handled, entity))
		is_thing = self._is_thing
		handled_char_thing = self._handled_char_thing
		for (
			charn, thingn, rulebook, rulen
		) in self._character_thing_rules_handled_cache.iter_unhandled_rules(
			branch, turn, tick):
			if not node_exists(charn, thingn) or not is_thing(charn, thingn):
				continue
			self.debug(
				f"checking triggers for character-thing rule: {charn, thingn, rulebook, rulen}"
			)
			rule = rulemap[rulen]
			handled = partial(handled_char_thing, charn, thingn, rulebook,
								rulen, branch, turn)
			entity = get_node(charn, thingn)
			if check_triggers(rule, handled, entity):
				self.debug(
					f"character-thing rule triggered: {charn, thingn, rulebook, rulen}"
				)
				todo[rulebook].append((rule, handled, entity))
		handled_char_place = self._handled_char_place
		for (
			charn, placen, rulebook, rulen
		) in self._character_place_rules_handled_cache.iter_unhandled_rules(
			branch, turn, tick):
			if not node_exists(charn, placen) or is_thing(charn, placen):
				continue
			self.debug(
				f"checking triggers for character-place rule: {charn, placen, rulebook, rulen}"
			)
			rule = rulemap[rulen]
			handled = partial(handled_char_place, charn, placen, rulebook,
								rulen, branch, turn)
			entity = get_node(charn, placen)
			if check_triggers(rule, handled, entity):
				self.debug(
					f"character-place rule triggered: {charn, placen, rulebook, rulen}"
				)
				todo[rulebook].append((rule, handled, entity))
		edge_exists = self._edge_exists
		get_edge = self._get_edge
		handled_char_port = self._handled_char_port
		for (
			charn, orign, destn, rulebook, rulen
		) in self._character_portal_rules_handled_cache.iter_unhandled_rules(
			branch, turn, tick):
			if not edge_exists(charn, orign, destn):
				continue
			self.debug(
				f"checking triggers for character-portal rule: {charn, orign, destn, rulebook, rulen}"
			)
			rule = rulemap[rulen]
			handled = partial(handled_char_port, charn, orign, destn, rulebook,
								rulen, branch, turn)
			entity = get_edge(charn, orign, destn)
			if check_triggers(rule, handled, entity):
				self.debug(
					f"character-portal rule triggered: {charn, orign, destn, rulebook, rulen}"
				)
				todo[rulebook].append((rule, handled, entity))
		handled_node = self._handled_node
		for (charn, noden, rulebook,
				rulen) in self._node_rules_handled_cache.iter_unhandled_rules(
					branch, turn, tick):
			if not node_exists(charn, noden):
				continue
			self.debug(
				f"checking triggers for node rule: {charn, noden, rulebook, rulen}"
			)
			rule = rulemap[rulen]
			handled = partial(handled_node, charn, noden, rulebook, rulen,
								branch, turn)
			entity = get_node(charn, noden)
			if check_triggers(rule, handled, entity):
				self.debug(
					f"node rule triggered: {charn, noden, rulebook, rulen}")
				todo[rulebook].append((rule, handled, entity))
		handled_portal = self._handled_portal
		for (
			charn, orign, destn, rulebook,
			rulen) in self._portal_rules_handled_cache.iter_unhandled_rules(
				branch, turn, tick):
			if not edge_exists(charn, orign, destn):
				continue
			self.debug(
				f"checking triggers for portal rule: {charn, orign, destn, rulebook, rulen}"
			)
			rule = rulemap[rulen]
			handled = partial(handled_portal, charn, orign, destn, rulebook,
								rulen, branch, turn)
			entity = get_edge(charn, orign, destn)
			if check_triggers(rule, handled, entity):
				self.debug(
					f"portal rule triggered: {charn, orign, destn, rulebook, rulen}"
				)
				todo[rulebook].append((rule, handled, entity))

		# TODO: rulebook priorities (not individual rule priorities, just follow the order of the rulebook)
		def fmtent(entity):
			if isinstance(entity, self.char_cls):
				return entity.name
			elif hasattr(entity, 'name'):
				return f"{entity.character.name}.node[{entity.name}]"
			else:
				return f"{entity.character.name}.portal[{entity.origin.name}][{entity.destination.name}]"

		for rulebook in sort_set(todo.keys()):
			for rule, handled, entity in todo[rulebook]:
				if not entity:
					continue
				self.debug(
					f"checking prereqs for rule {rule} on entity {fmtent(entity)}"
				)
				if check_prereqs(rule, handled, entity):
					self.debug(
						f"prereqs for rule {rule} on entity {fmtent(entity)} satisfied, will run actions"
					)
					try:
						yield do_actions(rule, handled, entity)
						self.debug(
							f"actions for rule {rule} on entity {fmtent(entity)} have run without incident"
						)
					except StopIteration:
						raise InnerStopIteration

	def advance(self) -> Any:
		"""Follow the next rule if available.

        If we've run out of rules, reset the rules iterator.

        """
		assert self.turn > self._turns_completed[self.branch]
		try:
			return next(self._rules_iter)
		except InnerStopIteration:
			self._rules_iter = self._follow_rules()
			return StopIteration()
		except StopIteration:
			self._rules_iter = self._follow_rules()
			return final_rule

	# except Exception as ex:
	#     self._rules_iter = self._follow_rules()
	#     return ex

	def new_character(self,
						name: Hashable,
						data: Graph = None,
						**kwargs) -> Character:
		"""Create and return a new :class:`Character`."""
		self.add_character(name, data, **kwargs)
		return self.character[name]

	new_graph = new_character

	def add_character(self,
						name: Hashable,
						data: Graph = None,
						**kwargs) -> None:
		"""Create a new character.

        You'll be able to access it as a :class:`Character` object by
        looking up ``name`` in my ``character`` property.

        ``data``, if provided, should be a networkx-compatible graph
        object. Your new character will be a copy of it.

        Any keyword arguments will be set as stats of the new character.

        """
		self._init_graph(name, 'DiGraph', data)
		self._graph_objs[name] = graph_obj = self.char_cls(self, name)
		if kwargs:
			graph_obj.stat.update(kwargs)

	def del_character(self, name: Hashable) -> None:
		"""Remove the Character from the database entirely.

        This also deletes all its history. You'd better be sure.

        """
		self.query.del_character(name)
		self.del_graph(name)
		del self.character[name]

	def _is_thing(self, character: Hashable, node: Hashable) -> bool:
		return self._things_cache.contains_entity(character, node,
													*self._btt())

	def _set_thing_loc(self, character: Hashable, node: Hashable,
						loc: Hashable) -> None:
		branch, turn, tick = self._nbtt()
		self._things_cache.store(character, node, branch, turn, tick, loc)
		self.query.set_thing_loc(character, node, branch, turn, tick, loc)

	def alias(self, v: Any, stat: Hashable = 'dummy') -> EntityStatAccessor:
		"""Return a pointer to a value for use in historical queries.

        It will behave much as if you assigned the value to some entity
        and then used its ``historical`` method to get a reference to
        the set of its past values, which happens to contain only the
        value you've provided here, ``v``.

        :arg v: the value to represent
        :arg stat: what name to pretend its stat has; usually irrelevant

        """
		from .util import EntityStatAccessor
		r = DummyEntity(self)
		r[stat] = v
		return EntityStatAccessor(r, stat, engine=self)

	def _entityfy(self,
					v: Any,
					stat: Hashable = 'dummy') -> EntityStatAccessor:
		from .query import Query
		if (isinstance(v, self.thing_cls) or isinstance(v, self.place_cls)
			or isinstance(v, self.portal_cls) or isinstance(v, Query)
			or isinstance(v, EntityStatAccessor)):
			return v
		return self.alias(v, stat)

	def _snap_keyframe(self, graph: Hashable, branch: str, turn: int,
						tick: int, nodes: NodeValDictType,
						edges: EdgeValDictType,
						graph_val: StatDictType) -> None:
		super()._snap_keyframe(graph, branch, turn, tick, nodes, edges,
								graph_val)
		newkf = {}
		contkf = {}
		for (name, node) in nodes.items():
			if 'location' not in node:
				continue
			locn = node['location']
			newkf[name] = locn
			if locn in contkf:
				contkf[locn].add(name)
			else:
				contkf[locn] = {
					name,
				}
		contents_keyframes = self._node_contents_cache.keyframe
		contkfs = contents_keyframes[graph, ][branch]
		if turn not in contkfs:
			contkfs[turn] = {tick: contkf}
		else:
			contkfs[turn][tick] = contkf
		kfs = self._things_cache.keyframe[graph, ][branch]
		if turn not in kfs:
			kfs[turn] = {tick: newkf}
		else:
			kfs[turn][tick] = newkf

	def turns_when(self, qry):
		"""Yield the turns in this branch when the query held true

        :arg qry: a Query, likely constructed by comparing the result
        of a call to an entity's ``historical`` method with the output
        of ``self.alias(..)`` or another ``historical(..)``

        """
		# yeah, it's just a loop over the query's method...I'm planning
		# on moving some iter_turns logic in here when I figure out what
		# of it is truly independent of any given type of query
		for branch, turn in qry.iter_turns():
			yield turn

	def _node_contents(self, character: Hashable, node: Hashable) -> Set:
		return self._node_contents_cache.retrieve(character, node,
													*self._btt())

	def apply_choices(
		self,
		choices: List[dict],
		dry_run=False,
		perfectionist=False
	) -> Tuple[List[Tuple[Any, Any]], List[Tuple[Any, Any]]]:
		"""Validate changes a player wants to make, and apply if acceptable.

        Returns a pair of lists containing acceptance and rejection messages,
        which the UI may present as it sees fit. They are always in a pair with
        the change request as the zeroth item. The message may be None or a string.

        Validator functions may return only a boolean indicating acceptance.
        If they instead return a pair, the initial boolean indicates acceptance
        and the following item is the message.

        This function will not actually result in any simulation happening.
        It creates a plan. See my ``plan`` context manager for the precise
        meaning of this.

        With ``dry_run=True`` just return the acceptances and rejections without
        really planning anything. With ``perfectionist=True`` apply changes if
        and only if all of them are accepted.

        """
		schema = self.schema
		todo = defaultdict(list)
		acceptances = []
		rejections = []
		for track in choices:
			entity = track['entity']
			permissible = schema.entity_permitted(entity)
			if isinstance(permissible, tuple):
				permissible, msg = permissible
			else:
				msg = ''
			if not permissible:
				for turn, changes in enumerate(track['changes'],
												start=self.turn):
					rejections.extend(
						((turn, entity, k, v), msg) for (k, v) in changes)
				continue
			for turn, changes in enumerate(track['changes'], start=self.turn):
				for k, v in changes:
					ekv = (entity, k, v)
					parcel = (turn, entity, k, v)
					val = schema.stat_permitted(*parcel)
					if type(val) is tuple:
						accept, message = val
						if accept:
							todo[turn].append(ekv)
							l = acceptances
						else:
							l = rejections
						l.append((parcel, message))
					elif val:
						todo[turn].append(ekv)
						acceptances.append((parcel, None))
					else:
						rejections.append((parcel, None))
		if dry_run or (perfectionist and rejections):
			return acceptances, rejections
		now = self.turn
		with self.plan():
			for turn in sorted(todo):
				self.turn = turn
				for entity, key, value in todo[turn]:
					if isinstance(entity, self.char_cls):
						entity.stat[key] = value
					else:
						entity[key] = value
		self.turn = now
		return acceptances, rejections



# This file is part of allegedb, an object relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""SQLAlchemy code to generate the SQL used by the allegedb ORM

If SQLAlchemy is installed at runtime, this will use it to generate SQL on the fly;
if it is not, allegedb can use a pregenerated file "sqlite.json" to store graphs in
a SQLite database. Generate sqlite.json by running this on the command line:

python3 alchemy.py >sqlite.json

"""
from functools import partial
from sqlalchemy import (
	Table,
	Column,
	CheckConstraint,
	ForeignKeyConstraint,
	INT,
	TEXT,
	BOOLEAN,
	MetaData,
	ForeignKey,
	select,
	func,
)

BaseColumn = Column
Column = partial(BaseColumn, nullable=False)

from sqlalchemy.sql import bindparam, and_, or_
from json import dumps
from functools import partial


def tables_for_meta(meta):
	Table('global',
			meta,
			Column('key', TEXT, primary_key=True),
			Column('value', TEXT, nullable=True),
			sqlite_with_rowid=False)
	Table('branches',
			meta,
			Column('branch',
					TEXT,
					ForeignKey('branches.parent'),
					primary_key=True,
					default='trunk'),
			Column('parent', TEXT, default='trunk', nullable=True),
			Column('parent_turn', INT, default=0),
			Column('parent_tick', INT, default=0),
			Column('end_turn', INT, default=0),
			Column('end_tick', INT, default=0),
			CheckConstraint('branch<>parent'),
			sqlite_with_rowid=False)
	Table('turns',
			meta,
			Column('branch', TEXT, primary_key=True),
			Column('turn', INT, primary_key=True),
			Column('end_tick', INT),
			Column('plan_end_tick', INT),
			sqlite_with_rowid=False)
	Table('graphs',
			meta,
			Column('graph', TEXT, primary_key=True),
			Column('type', TEXT, default='Graph'),
			CheckConstraint(
				"type IN ('Graph', 'DiGraph', 'MultiGraph', 'MultiDiGraph')"),
			sqlite_with_rowid=False)
	Table('keyframes',
			meta,
			Column('graph', TEXT, ForeignKey('graphs.graph'),
					primary_key=True),
			Column('branch',
					TEXT,
					ForeignKey('branches.branch'),
					primary_key=True,
					default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('nodes', TEXT),
			Column('edges', TEXT),
			Column('graph_val', TEXT),
			sqlite_with_rowid=False)
	Table('graph_val',
			meta,
			Column('graph', TEXT, ForeignKey('graphs.graph'),
					primary_key=True),
			Column('key', TEXT, primary_key=True),
			Column('branch',
					TEXT,
					ForeignKey('branches.branch'),
					primary_key=True,
					default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('value', TEXT, nullable=True),
			sqlite_with_rowid=False)
	Table('nodes',
			meta,
			Column('graph', TEXT, ForeignKey('graphs.graph'),
					primary_key=True),
			Column('node', TEXT, primary_key=True),
			Column('branch',
					TEXT,
					ForeignKey('branches.branch'),
					primary_key=True,
					default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('extant', BOOLEAN),
			sqlite_with_rowid=False)
	Table('node_val',
			meta,
			Column('graph', TEXT, primary_key=True),
			Column('node', TEXT, primary_key=True),
			Column('key', TEXT, primary_key=True),
			Column('branch',
					TEXT,
					ForeignKey('branches.branch'),
					primary_key=True,
					default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('value', TEXT, nullable=True),
			ForeignKeyConstraint(['graph', 'node'],
									['nodes.graph', 'nodes.node']),
			sqlite_with_rowid=False)
	Table('edges',
			meta,
			Column('graph', TEXT, ForeignKey('graphs.graph'),
					primary_key=True),
			Column('orig', TEXT, primary_key=True),
			Column('dest', TEXT, primary_key=True),
			Column('idx', INT, primary_key=True),
			Column('branch',
					TEXT,
					ForeignKey('branches.branch'),
					primary_key=True,
					default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('extant', BOOLEAN),
			ForeignKeyConstraint(['graph', 'orig'],
									['nodes.graph', 'nodes.node']),
			ForeignKeyConstraint(['graph', 'dest'],
									['nodes.graph', 'nodes.node']),
			sqlite_with_rowid=False)
	Table('edge_val',
			meta,
			Column('graph', TEXT, primary_key=True),
			Column('orig', TEXT, primary_key=True),
			Column('dest', TEXT, primary_key=True),
			Column('idx', INT, primary_key=True),
			Column('key', TEXT, primary_key=True),
			Column('branch',
					TEXT,
					ForeignKey('branches.branch'),
					primary_key=True,
					default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('value', TEXT, nullable=True),
			ForeignKeyConstraint(
				['graph', 'orig', 'dest', 'idx'],
				['edges.graph', 'edges.orig', 'edges.dest', 'edges.idx']),
			sqlite_with_rowid=False)
	Table('plans', meta, Column('id', INT, primary_key=True),
			Column('branch', TEXT), Column('turn', INT), Column('tick', INT))
	Table('plan_ticks',
			meta,
			Column('plan_id', INT, primary_key=True),
			Column('turn', INT, primary_key=True),
			Column('tick', INT, primary_key=True),
			ForeignKeyConstraint(('plan_id', ), ('plans.id', )),
			sqlite_with_rowid=False)
	return meta.tables


def indices_for_table_dict(table):
	return {}


def queries_for_table_dict(table):

	def tick_to_end_clause(tab):
		return and_(
			tab.c.graph == bindparam('graph'),
			tab.c.branch == bindparam('branch'),
			or_(
				tab.c.turn > bindparam('turn_from_a'),
				and_(tab.c.turn == bindparam('turn_from_b'),
						tab.c.tick >= bindparam('tick_from'))))

	def tick_to_tick_clause(tab):
		return and_(
			tick_to_end_clause(tab),
			or_(
				tab.c.turn < bindparam('turn_to_a'),
				and_(tab.c.turn == bindparam('turn_to_b'),
						tab.c.tick <= bindparam('tick_to'))))

	r = {
		'global_get':
		select([table['global'].c.value
				]).where(table['global'].c.key == bindparam('key')),
		'global_update':
		table['global'].update().values(value=bindparam('value')).where(
			table['global'].c.key == bindparam('key')),
		'graph_type':
		select([table['graphs'].c.type
				]).where(table['graphs'].c.graph == bindparam('graph')),
		'del_edge_val_graph':
		table['edge_val'].delete().where(
			table['edge_val'].c.graph == bindparam('graph')),
		'del_edge_val_after':
		table['edge_val'].delete().where(
			and_(
				table['edge_val'].c.graph == bindparam('graph'),
				table['edge_val'].c.orig == bindparam('orig'),
				table['edge_val'].c.dest == bindparam('dest'),
				table['edge_val'].c.idx == bindparam('idx'),
				table['edge_val'].c.key == bindparam('key'),
				table['edge_val'].c.branch == bindparam('branch'),
				or_(
					table['edge_val'].c.turn > bindparam('turn'),
					and_(table['edge_val'].c.turn == bindparam('turn'),
							table['edge_val'].c.tick >= bindparam('tick'))))),
		'del_edges_graph':
		table['edges'].delete().where(
			table['edges'].c.graph == bindparam('graph')),
		'del_edges_after':
		table['edges'].delete().where(
			and_(
				table['edges'].c.graph == bindparam('graph'),
				table['edges'].c.orig == bindparam('orig'),
				table['edges'].c.dest == bindparam('dest'),
				table['edges'].c.idx == bindparam('idx'),
				table['edges'].c.branch == bindparam('branch'),
				or_(
					table['edges'].c.turn > bindparam('turn'),
					and_(table['edges'].c.turn == bindparam('turn'),
							table['edges'].c.tick >= bindparam('tick'))))),
		'del_nodes_graph':
		table['nodes'].delete().where(
			table['nodes'].c.graph == bindparam('graph')),
		'del_nodes_after':
		table['nodes'].delete().where(
			and_(
				table['nodes'].c.graph == bindparam('graph'),
				table['nodes'].c.node == bindparam('node'),
				table['nodes'].c.branch == bindparam('branch'),
				or_(
					table['nodes'].c.turn > bindparam('turn'),
					and_(table['nodes'].c.turn == bindparam('turn'),
							table['nodes'].c.tick >= bindparam('tick'))))),
		'del_node_val_graph':
		table['node_val'].delete().where(
			table['node_val'].c.graph == bindparam('graph')),
		'del_node_val_after':
		table['node_val'].delete().where(
			and_(
				table['node_val'].c.graph == bindparam('graph'),
				table['node_val'].c.node == bindparam('node'),
				table['node_val'].c.key == bindparam('key'),
				table['node_val'].c.branch == bindparam('branch'),
				or_(
					table['node_val'].c.turn > bindparam('turn'),
					and_(table['node_val'].c.turn == bindparam('turn'),
							table['node_val'].c.tick >= bindparam('tick'))))),
		'del_graph':
		table['graphs'].delete().where(
			table['graphs'].c.graph == bindparam('graph')),
		'del_graph_val_after':
		table['graph_val'].delete().where(
			and_(
				table['graph_val'].c.graph == bindparam('graph'),
				table['graph_val'].c.key == bindparam('key'),
				table['graph_val'].c.branch == bindparam('branch'),
				or_(
					table['graph_val'].c.turn > bindparam('turn'),
					and_(table['graph_val'].c.turn == bindparam('turn'),
							table['graph_val'].c.tick >= bindparam('tick'))))),
		'global_delete':
		table['global'].delete().where(
			table['global'].c.key == bindparam('key')),
		'graphs_types':
		select([table['graphs'].c.graph, table['graphs'].c.type]),
		'graphs_named':
		select([func.COUNT()]).select_from(table['graphs']).where(
			table['graphs'].c.graph == bindparam('graph')),
		'update_branches':
		table['branches'].update().values(
			parent=bindparam('parent'),
			parent_turn=bindparam('parent_turn'),
			parent_tick=bindparam('parent_tick'),
			end_turn=bindparam('end_turn'),
			end_tick=bindparam('end_tick')).where(
				table['branches'].c.branch == bindparam('branch')),
		'update_turns':
		table['turns'].update().values(
			end_tick=bindparam('end_tick'),
			plan_end_tick=bindparam('plan_end_tick')).where(
				and_(table['turns'].c.branch == bindparam('branch'),
						table['turns'].c.turn == bindparam('turn'))),
		'keyframes_list':
		select([
			table['keyframes'].c.graph, table['keyframes'].c.branch,
			table['keyframes'].c.turn, table['keyframes'].c.tick
		]),
		'get_keyframe':
		select([
			table['keyframes'].c.nodes, table['keyframes'].c.edges,
			table['keyframes'].c.graph_val
		]).where(
			and_(table['keyframes'].c.graph == bindparam('graph'),
					table['keyframes'].c.branch == bindparam('branch'),
					table['keyframes'].c.turn == bindparam('turn'),
					table['keyframes'].c.tick == bindparam('tick'))),
		'load_nodes_tick_to_end':
		select([
			table['nodes'].c.node, table['nodes'].c.turn,
			table['nodes'].c.tick, table['nodes'].c.extant
		]).where(tick_to_end_clause(table['nodes'])),
		'load_nodes_tick_to_tick':
		select([
			table['nodes'].c.node, table['nodes'].c.turn,
			table['nodes'].c.tick, table['nodes'].c.extant
		]).where(tick_to_tick_clause(table['nodes'])),
		'load_edges_tick_to_end':
		select([
			table['edges'].c.orig, table['edges'].c.dest, table['edges'].c.idx,
			table['edges'].c.turn, table['edges'].c.tick,
			table['edges'].c.extant
		]).where(tick_to_end_clause(table['edges'])),
		'load_edges_tick_to_tick':
		select([
			table['edges'].c.orig, table['edges'].c.dest, table['edges'].c.idx,
			table['edges'].c.turn, table['edges'].c.tick,
			table['edges'].c.extant
		]).where(tick_to_tick_clause(table['edges'])),
		'load_node_val_tick_to_end':
		select([
			table['node_val'].c.node, table['node_val'].c.key,
			table['node_val'].c.turn, table['node_val'].c.tick,
			table['node_val'].c.value
		]).where(tick_to_end_clause(table['node_val'])),
		'load_node_val_tick_to_tick':
		select([
			table['node_val'].c.node, table['node_val'].c.key,
			table['node_val'].c.turn, table['node_val'].c.tick,
			table['node_val'].c.value
		]).where(tick_to_tick_clause(table['node_val'])),
		'load_edge_val_tick_to_end':
		select([
			table['edge_val'].c.orig, table['edge_val'].c.dest,
			table['edge_val'].c.idx, table['edge_val'].c.key,
			table['edge_val'].c.turn, table['edge_val'].c.tick,
			table['edge_val'].c.value
		]).where(tick_to_end_clause(table['edge_val'])),
		'load_edge_val_tick_to_tick':
		select([
			table['edge_val'].c.orig, table['edge_val'].c.dest,
			table['edge_val'].c.idx, table['edge_val'].c.key,
			table['edge_val'].c.turn, table['edge_val'].c.tick,
			table['edge_val'].c.value
		]).where(tick_to_tick_clause(table['edge_val'])),
		'load_graph_val_tick_to_end':
		select([
			table['graph_val'].c.key, table['graph_val'].c.turn,
			table['graph_val'].c.tick, table['graph_val'].c.value
		]).where(tick_to_end_clause(table['graph_val'])),
		'load_graph_val_tick_to_tick':
		select([
			table['graph_val'].c.key, table['graph_val'].c.turn,
			table['graph_val'].c.tick, table['graph_val'].c.value
		]).where(tick_to_tick_clause(table['graph_val']))
	}
	for t in table.values():
		key = list(t.primary_key)
		if 'branch' in t.columns and 'turn' in t.columns and 'tick' in t.columns:
			branch = t.columns['branch']
			turn = t.columns['turn']
			tick = t.columns['tick']
			if branch in key and turn in key and tick in key:
				key = [branch, turn, tick]
				r[t.name + '_del_time'] = t.delete().where(
					and_(t.c.branch == bindparam('branch'),
							t.c.turn == bindparam('turn'),
							t.c.tick == bindparam('tick')))
		r[t.name + '_dump'] = select(list(t.c.values())).order_by(*key)
		r[t.name + '_insert'] = t.insert().values(
			tuple(bindparam(cname) for cname in t.c.keys()))
		r[t.name + '_count'] = select([func.COUNT()]).select_from(t)
		r[t.name + '_del'] = t.delete().where(
			and_(*[c == bindparam(c.name) for c in t.primary_key]))
	return r


def gather_sql(meta):
	from sqlalchemy.sql.ddl import CreateTable, CreateIndex
	r = {}
	table = tables_for_meta(meta)
	index = indices_for_table_dict(table)
	query = queries_for_table_dict(table)

	for t in table.values():
		r['create_' + t.name] = CreateTable(t)
	for (tab, idx) in index.items():
		r['index_' + tab] = CreateIndex(idx)
	r.update(query)

	return r


def compile_sql(dialect, meta):
	return {
		k: v.compile(dialect=dialect)
		for (k, v) in gather_sql(meta).items()
	}


class Alchemist(object):
	"""Holds an engine and runs queries on it.

    """

	def __init__(self, engine):
		self.engine = engine
		self.conn = self.engine.connect()
		self.meta = MetaData()
		self.sql = gather_sql(self.meta)

		def caller(k, *largs):
			statement = self.sql[k].compile(dialect=self.conn.engine.dialect)
			if hasattr(statement, 'positiontup'):
				return self.conn.execute(
					statement, **dict(zip(statement.positiontup, largs)))
			elif largs:
				raise TypeError("{} is a DDL query, I think".format(k))
			return self.conn.execute(self.sql[k])

		def manycaller(k, *largs):
			statement = self.sql[k].compile(dialect=self.conn.engine.dialect)
			return self.conn.execute(
				statement,
				*(dict(zip(statement.positiontup, larg)) for larg in largs))

		class Many(object):
			pass

		self.many = Many()
		for (key, query) in self.sql.items():
			setattr(self, key, partial(caller, key))
			setattr(self.many, key, partial(manycaller, key))


if __name__ == '__main__':
	from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite

	out = dict((k, str(v)) for (
		k, v) in compile_sql(SQLiteDialect_pysqlite(), MetaData()).items())

	print(dumps(out, indent=4, sort_keys=True))



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from LiSE.proxy import EngineProcessManager
import LiSE.allegedb.tests.test_all
import pytest
import LiSE.examples.kobold as kobold
import LiSE.examples.college as college
import shutil
import tempfile


class ProxyTest(LiSE.allegedb.tests.test_all.AllegedTest):

	def setUp(self):
		self.manager = EngineProcessManager()
		self.tempdir = tempfile.mkdtemp(dir='.')
		self.engine = self.manager.start(
			self.tempdir, connect_string='sqlite:///:memory:')
		self.graphmakers = (self.engine.new_character, )
		self.addCleanup(self._do_cleanup)

	def _do_cleanup(self):
		self.manager.shutdown()
		shutil.rmtree(self.tempdir)


class ProxyGraphTest(LiSE.allegedb.tests.test_all.AbstractGraphTest,
						ProxyTest):
	pass


class DictStorageTest(ProxyTest, LiSE.allegedb.tests.test_all.DictStorageTest):
	pass


class ListStorageTest(ProxyTest, LiSE.allegedb.tests.test_all.ListStorageTest):
	pass


class SetStorageTest(ProxyTest, LiSE.allegedb.tests.test_all.SetStorageTest):
	pass


@pytest.fixture(scope='function')
def handle(tempdir):
	from LiSE.handle import EngineHandle
	hand = EngineHandle((tempdir, ), {
		'connect_string': 'sqlite:///:memory:',
		'random_seed': 69105
	})
	yield hand
	hand.close()


@pytest.fixture(
	scope='function',
	params=[
		lambda eng: kobold.inittest(
			eng, shrubberies=20, kobold_sprint_chance=.9),
		# college.install,
		# sickle.install
	])
def handle_initialized(request, handle):
	with handle._real.advancing():
		request.param(handle._real)
	yield handle


def test_fast_delta(handle_initialized):
	hand = handle_initialized

	# there's currently no way to do fast delta past the time when
	# a character was created, due to the way keyframes work...
	# so don't test that
	def unpack_delta(d):
		catted = hand._concat_char_delta(d)
		assert isinstance(catted, bytes)
		return hand.unpack(catted)

	branch, turn, tick = hand._real._btt()
	ret, diff = hand.next_turn()
	btt = hand._real._btt()
	slowd = unpack_delta(
		hand._get_slow_delta(btt_from=(branch, turn, tick), btt_to=btt))
	assert hand.unpack(diff) == slowd, "Fast delta differs from slow delta"
	ret, diff2 = hand.time_travel('trunk', 0, tick)
	btt2 = hand._real._btt()
	slowd2 = unpack_delta(hand._get_slow_delta(btt_from=btt, btt_to=btt2))
	assert hand.unpack(diff2) == slowd2, "Fast delta differs from slow delta"
	ret, diff3 = hand.time_travel('trunk', 3)
	btt3 = hand._real._btt()
	slowd3 = unpack_delta(hand._get_slow_delta(btt_from=btt2, btt_to=btt3))
	assert hand.unpack(diff3) == slowd3, "Fast delta differs from slow delta"
	ret, diff4 = hand.time_travel('trunk', 1)
	btt4 = hand._real._btt()
	slowd4 = unpack_delta(hand._get_slow_delta(btt_from=btt3, btt_to=btt4))
	assert hand.unpack(diff4) == slowd4, "Fast delta differs from slow delta"


@pytest.mark.slow
def test_serialize_deleted(engy):
	eng = engy
	with eng.advancing():
		college.install(eng)
	d0r0s0 = eng.character['dorm0room0student0']
	roommate = d0r0s0.stat['roommate']
	del eng.character[roommate.name]
	assert not roommate
	with pytest.raises(KeyError):
		eng.character[roommate.name]
	assert d0r0s0.stat['roommate'] == roommate
	assert eng.unpack(eng.pack(d0r0s0.stat['roommate'])) == roommate


def test_manip_deleted(engy):
	eng = engy
	phys = eng.new_character('physical')
	phys.stat['aoeu'] = True
	phys.add_node(0)
	phys.add_node(1)
	phys.node[1]['aoeu'] = True
	del phys.node[1]
	phys.add_node(1)
	assert 'aoeu' not in phys.node[1]
	phys.add_edge(0, 1)
	phys.adj[0][1]['aoeu'] = True
	del phys.adj[0][1]
	phys.add_edge(0, 1)
	assert 'aoeu' not in phys.adj[0][1]
	del eng.character['physical']
	assert not phys
	phys = eng.new_character('physical')
	assert 'aoeu' not in phys.stat
	assert 0 not in phys
	assert 1 not in phys
	assert 0 not in phys.adj
	assert 1 not in phys.adj



import pytest
from LiSE import allegedb
import os


@pytest.fixture(scope='function')
def orm():
	with allegedb.ORM("sqlite:///:memory:") as it:
		yield it


def test_single_plan(orm):
	g = orm.new_digraph('graph')
	g.add_node(0)
	orm.turn = 1
	g.add_node(1)
	with orm.plan():
		orm.turn = 2
		g.add_node(2)
		g.node[2]['clever'] = False
		orm.turn = 3
		g.node[2]['funny'] = True
		g.add_node(3)
		orm.turn = 4
		g.node[2]['successful'] = True
	orm.turn = 1
	assert 2 not in g.node
	orm.branch = 'b'
	assert 2 not in g.node
	assert 1 in g
	orm.turn = 2
	assert 2 in g.node
	assert g.node[2].keys() == {'clever'}
	orm.turn = 3
	assert g.node[2]['funny']
	assert 3 in g
	assert g.node[2].keys() == {'funny', 'clever'}
	orm.turn = 4
	assert g.node[2].keys() == {'funny', 'clever', 'successful'}
	orm.turn = 2
	assert g.node[2].keys() == {'clever'}
	g.node[2]['funny'] = False
	assert g.node[2].keys() == {'funny', 'clever'}
	orm.turn = 3
	assert not g.node[2]['funny']
	assert 3 not in g.node
	orm.turn = 4
	assert g.node[2].keys() == {'funny', 'clever'}
	orm.turn = 1
	orm.branch = 'trunk'
	orm.turn = 0
	assert 1 not in g.node
	orm.branch = 'c'
	orm.turn = 2
	assert 1 not in g.node
	assert 2 not in g.node
	orm.turn = 0
	orm.branch = 'trunk'
	orm.turn = 2
	assert 2 in g.node


def test_multi_plan(orm):
	g1 = orm.new_digraph(1)
	g2 = orm.new_digraph(2)
	with orm.plan():
		g1.add_node(1)
		g1.add_node(2)
		orm.turn = 1
		g1.add_edge(1, 2)
	orm.turn = 0
	with orm.plan():
		g2.add_node(1)
		g2.add_node(2)
		orm.turn = 1
		g2.add_edge(1, 2)
	orm.turn = 0
	# contradict the plan
	del g1.node[2]
	assert 1 in g2.node
	assert 2 in g2.node
	orm.turn = 1
	assert 2 not in g1.node
	assert 2 not in g1.edge[1]
	assert 2 in g2.edge[1]


def test_plan_vs_plan(orm):
	g1 = orm.new_digraph(1)
	with orm.plan():
		g1.add_node(1)
		g1.add_node(2)
		orm.turn = 1
		g1.add_edge(1, 2)
		g1.add_node(3)
		g1.add_edge(3, 1)
	orm.turn = 0
	with orm.plan():
		g1.add_node(0)  # not a contradiction, just two plans
		g1.add_edge(0, 1)
	orm.turn = 1
	assert 0 in g1.node
	assert 1 in g1.node
	assert 2 in g1.node
	assert 3 in g1.node
	assert 1 in g1.edge[0]
	assert 2 in g1.edge[1]
	orm.turn = 0
	with orm.plan():
		del g1.node[2]
	orm.turn = 0  # go to turn end
	assert 2 not in g1.node
	orm.turn = 2
	assert 3 not in g1.node
	assert 3 not in g1.adj
	assert 0 in g1.node
	assert 1 in g1.adj[0]


def test_save_load_plan(tmpdbfile):
	tmpdbstring = 'sqlite:///' + tmpdbfile
	with allegedb.ORM(tmpdbstring) as orm:
		g1 = orm.new_digraph(1)
		g2 = orm.new_digraph(2)
		with orm.plan():
			g1.add_node(1)
			g1.add_node(2)
			orm.turn = 1
			g1.add_edge(1, 2)
		orm.turn = 0
		with orm.plan():
			g2.add_node(1)
			g2.add_node(2)
			orm.turn = 1
			g2.add_edge(1, 2)
		orm.turn = 0
	with allegedb.ORM(tmpdbstring) as orm:
		g1 = orm.graph[1]
		g2 = orm.graph[2]
		# go to end of turn
		orm.turn = 0
		# contradict the plan
		del g1.node[2]
		assert 1 in g2.node
		assert 2 in g2.node
		orm.turn = 1
		assert 2 not in g1.node
		assert 2 not in g1.edge[1]
		assert 2 in g2.edge[1]
	with allegedb.ORM(tmpdbstring) as orm:
		g1 = orm.graph[1]
		g2 = orm.graph[2]
		assert 1 in g2.node
		assert 2 in g2.node
		orm.turn = 1
		assert 2 not in g1.node
		assert 2 not in g1.edge[1]
		assert 2 in g2.edge[1]
		orm.turn = 0
		while 2 in g2.node:
			orm.tick -= 1
		g2.add_node(2)  # contradict plan
		orm.turn = 0  # end of turn
		assert 2 not in g2.edge[1]
